#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; A fixnum-centric canvas.
;;; package: (gtk fix-layout)

;;; <fix-widget> is the base class that handles the realization of a
;;; widget's GdkWindow.  It will allocate/move/resize the GdkWindow
;;; and dispatch events received on it.

(define-class (<fix-widget> (constructor () (width height bgcolor)))
    (<scm-widget>)

  ;; Our window -- a GdkWindow alien.  Until realized, a NULL pointer.
  (window define accessor
	  initializer (lambda () (make-alien '|GdkWindow|)))
  (%background-color define standard)

  ;; Our window geometry (allocation) -- a rectangular extent in
  ;; fixnum device coordinates (e.g. size in pixels, offset within
  ;; parent window [ancestor widget]).
  (geometry define accessor initializer (lambda () (make-fix-rect)))

  (event-handlers define accessor initializer
		  (lambda () (make-vector (C-enum "GDK_EVENT_LAST") #f))))

(define-guarantee fix-widget "a <fix-widget>")

(define-integrable (guarantee-size object operator)
  (guarantee non-negative-fixnum? object operator))

(define-method initialize-instance ((widget <fix-widget>) width height bgcolor)
  (let ((bg (if (null? bgcolor)
		'()
		(->color bgcolor '(initialize-instance <fix-widget>)))))
    (call-next-method widget)
    (%trace "; (initialize-instance <fix-widget>) "widget" "width"x"height"\n")
    (assert-glib-locked '(initialize-instance <fix-widget>))
    (set-scm-widget-natural-size! widget width height)
    (set-fix-widget-%background-color! widget bg)
    ;; Init. size, for a realize signal arriving before an allocation.
    (set-fix-rect-size! (fix-widget-geometry widget) width height)
    (let ((gtkwidget (gobject-alien widget)))
      (C-call "gtk_widget_set_has_window" gtkwidget 1)
      ;;(C-call "gtk_widget_set_app_paintable" gtkwidget 1)
      (C-call "gtk_widget_set_events" gtkwidget event-mask))

    (set-gtk-widget-realize-callback! widget fix-widget-realize-callback)
    (set-gtk-widget-unrealize-callback! widget fix-widget-unrealize-callback)
    (set-gtk-widget-size-allocate-callback! widget allocate-callback)
    (set-gtk-widget-event-callback! widget event-callback)))

(define-generic fix-widget-realize-callback (widget))

(define-method fix-widget-realize-callback ((widget <fix-widget>))
  (%trace "; (fix-widget-realize-callback <fix-widget>) "widget"\n")
  (assert-glib-locked '(fix-widget-realize-callback <fix-widget>))
  (let ((geometry (fix-widget-geometry widget))
	(attr (malloc (C-sizeof "GdkWindowAttr") '|GdkWindowAttr|))
	(main-GdkWindow (fix-widget-window widget))
	(parent-GdkWindow (make-alien '|GdkWindow|))
	(GtkWidget (gobject-alien widget)))

    ;; Create widget window.
    (C->= attr "GdkWindowAttr window_type" (C-enum "GDK_WINDOW_CHILD"))
    (C->= attr "GdkWindowAttr wclass" (C-enum "GDK_INPUT_OUTPUT"))
    (let ((x (fix-rect-x geometry))
	  (y (fix-rect-y geometry))
	  (width (fix-rect-width geometry))
	  (height (fix-rect-height geometry)))
      (if x (C->= attr "GdkWindowAttr x" x))
      (if y (C->= attr "GdkWindowAttr y" y))
      (C->= attr "GdkWindowAttr width" width)
      (C->= attr "GdkWindowAttr height" height)
      (C->= attr "GdkWindowAttr event_mask" event-mask)

      (C-call "gtk_widget_get_parent_window" parent-GdkWindow GtkWidget)
      (error-if-null parent-GdkWindow "Could not get parent:" widget)

      (C-call "gdk_window_new" main-GdkWindow parent-GdkWindow attr
	      (bit-ior (if x (C-enum "GDK_WA_X") 0)
		       (if y (C-enum "GDK_WA_Y") 0)))
      (free attr)
      (error-if-null main-GdkWindow "Could not create main window:" widget)
      (C-call "gtk_widget_set_window" GtkWidget main-GdkWindow)
      (C-call "gtk_widget_register_window" GtkWidget main-GdkWindow)
      ;; gtk_widget_unregister_window (as well as gdk_window_destroy)
      ;; are called by the default unrealize method.
      (%trace ";  window: "main-GdkWindow"\n"))))

(define event-mask
  (bit-ior
   (C-enum "GDK_ENTER_NOTIFY_MASK")
   (C-enum "GDK_LEAVE_NOTIFY_MASK")
   (C-enum "GDK_FOCUS_CHANGE_MASK")
   (C-enum "GDK_VISIBILITY_NOTIFY_MASK")
   (C-enum "GDK_KEY_PRESS_MASK")
   (C-enum "GDK_POINTER_MOTION_MASK")
   (C-enum "GDK_BUTTON_PRESS_MASK")
   (C-enum "GDK_BUTTON_RELEASE_MASK")))

(define-generic fix-widget-unrealize-callback (widget))

(define-method fix-widget-unrealize-callback ((widget <fix-widget>))
  (%trace "; (fix-widget-unrealize-callback <fix-widget>) "widget"\n")
  )

(define (allocate-callback widget GtkAllocation)
  (let ((x (C-> GtkAllocation "GtkAllocation x"))
	(y (C-> GtkAllocation "GtkAllocation y"))
	(width (C-> GtkAllocation "GtkAllocation width"))
	(height (C-> GtkAllocation "GtkAllocation height"))
	(rect (fix-widget-geometry widget)))
    (%trace "; allocated "width"x"height" at "x","y" for "widget"\n")
    (assert-glib-locked 'allocate-callback)
    (C-call "gtk_widget_set_allocation" (gobject-alien widget) GtkAllocation)
    (set-fix-rect! rect x y width height)
    (if (fix-widget-realized? widget)
	(C-call "gdk_window_move_resize"
		(fix-widget-window widget)
		x y width height))
    (fix-widget-new-geometry-callback widget)))

(define-generic fix-widget-new-geometry-callback (widget))

(define-method fix-widget-new-geometry-callback ((widget <fix-widget>))
  (declare (ignore widget))
  unspecific)

(define (fix-widget-realized? widget)
  (not (alien-null? (fix-widget-window widget))))

(define-syntax pointer-shapes
  (sc-macro-transformer
   (lambda (form usage-env)

     (define (simplify name)
       ;; |GDK_BASED_ARROW_DOWN| => based-arrow-down
       (let ((string (symbol->string name)))
	 (if (string-prefix? "GDK_" string)
	     (intern (string-replace (string-tail string 4) #\_ #\-))
	     (begin
	       (warn "Unexpected GdkCursorType name:" name)
	       name))))

     (list 'quote
	   (map (lambda (name.value)
		  (cons (simplify (car name.value))
			(cdr name.value)))
		(c-enum-constant-values '|GdkCursorType| form
					(find-c-includes usage-env)))))))

(define set-fix-widget-pointer-shape!
  (let ((alist (pointer-shapes)))
    (named-lambda (set-fix-widget-pointer-shape! widget name)
      (let ((name.value (or (assq name alist)
			    (error "Not a pointer shape:" name
				   (map car alist))))
	    (cursor (make-alien '|GdkCursor|))
	    (display (make-alien '|GdkDisplay|)))
	(assert-glib-locked 'set-fix-widget-pointer-shape!)
	;; GC-protect cursor!
	(C-call "gtk_widget_get_display" display (gobject-alien widget))
	(C-call "gdk_cursor_new_for_display" cursor display (cdr name.value))
	(C-call "gdk_window_set_cursor" (fix-widget-window widget) cursor)
	(C-call "g_object_unref" cursor)))))

(define (event-callback widget GdkEvent)
  (%trace2 ";event-callback "widget)

  (let ((type (C-> GdkEvent "GdkEvent any type")))
    (%trace2 " "(C-enum "GdkEventType" type)"\n")
    (let ((handler (vector-ref (fix-widget-event-handlers widget) type)))
      (and handler
	   (handler widget GdkEvent)))))

(define (set-fix-widget-enter-notify-handler! widget handler)
  (guarantee-fix-widget widget 'set-fix-widget-enter-notify-handler!)
  (guarantee-procedure-of-arity handler 1 'set-fix-widget-enter-notify-handler!)
  (vector-set!
   (fix-widget-event-handlers widget) (C-enum "GDK_ENTER_NOTIFY")
   (named-lambda (fix-widget-enter-notify-handler widget GdkEvent)
     (declare (ignore GdkEvent))
     (handler widget))))

(define (set-fix-widget-leave-notify-handler! widget handler)
  (guarantee-fix-widget widget 'set-fix-widget-leave-notify-handler!)
  (guarantee-procedure-of-arity handler 1 'set-fix-widget-leave-notify-handler!)
  (vector-set!
   (fix-widget-event-handlers widget) (C-enum "GDK_LEAVE_NOTIFY")
   (named-lambda (fix-widget-leave-notify-handler widget GdkEvent)
     (declare (ignore GdkEvent))
     (handler widget))))

(define (set-fix-widget-focus-change-handler! widget handler)
  (guarantee-fix-widget widget 'set-fix-widget-focus-change-handler!)
  (guarantee-procedure-of-arity handler 2 'set-fix-widget-focus-change-handler!)
  (vector-set!
   (fix-widget-event-handlers widget) (C-enum "GDK_FOCUS_CHANGE")
   (named-lambda (fix-widget-focus-change-handler widget GdkEvent)
     (let ((in? (not (zero? (C-> GdkEvent "GdkEventFocus in")))))
       (handler widget in?)))))

(define (set-fix-widget-visibility-notify-handler! widget handler)
  (guarantee-fix-widget widget 'set-fix-widget-visibility-notify-handler!)
  (guarantee-procedure-of-arity handler 2 'set-fix-widget-visibility-notify-handler!)
  (vector-set!
   (fix-widget-event-handlers widget) (C-enum "GDK_VISIBILITY_NOTIFY")
   (named-lambda (fix-widget-visibility-notify-handler widget GdkEvent)
     (let ((state (C-> GdkEvent "GdkEventVisibility state")))
       (handler
	widget
	(cond
	 ((int:= state (C-enum "GDK_VISIBILITY_UNOBSCURED")) 'VISIBLE)
	 ((int:= state (C-enum "GDK_VISIBILITY_PARTIAL")) 'PARTIALLY-OBSCURED)
	 ((int:= state (C-enum "GDK_VISIBILITY_FULLY_OBSCURED")) 'OBSCURED)
	 (else (C-enum "GdkVisibilityState" state))))))))

(define (set-fix-widget-key-press-handler! widget handler)
  (guarantee-fix-widget widget 'set-fix-widget-key-press-handler!)
  (guarantee-procedure-of-arity handler 3 'set-fix-widget-key-press-handler!)
  (vector-set!
   (fix-widget-event-handlers widget) (C-enum "GDK_KEY_PRESS")
   (named-lambda (fix-widget-key-press-handler widget GdkEvent)
     (let ((alien (C-> GdkEvent "GdkEvent key string"))
	   (length (C-> GdkEvent "GdkEvent key length"))
	   (state (C-> GdkEvent "GdkEvent key state"))
	   (keyval (C-> GdkEvent "GdkEvent key keyval")))
       (let ((string (c-peek-cstring alien))
	     (char-bits (gdk-key-state->char-bits state)))
	 (cond ((zero? (string-length string))
		(cond ((fix:= length 1)
		       (handler widget #\NUL char-bits))
		      ((fix:= length 0)
		       (handler widget (gdk-keyval->name keyval) char-bits))
		      (else (error "Unexpected length in GdkEventKey."))))
	       ;; Kludge: BackSpace and C-h both have "key string" "\b"?!
	       ;; And Delete is already "\177" (aka (string #\rubout)).
	       ((and (fix:= 1 (string-length string))
		     (char=? #\backspace (string-ref string 0)))
		(let ((name (gdk-keyval->name keyval)))
		  (cond ((string-ci=? (symbol->string name) "backspace")
			 (handler widget #\backspace char-bits))
			((memq name '(|h| |H|))
			 (handler widget #\C-h
				  (fix:- char-bits char-bit:control)))
			(else (error "Unexpected backspace keyval:" keyval)))))
	       (else
		(let ((l (string-length string)))
		  (let loop ((i 0))
		    (if (fix:< i l)
			(and (handler widget (string-ref string i) char-bits)
			     (loop (fix:1+ i)))
			#t))))))))))

(define (set-fix-widget-motion-handler! widget handler)
  (guarantee-fix-widget widget 'set-fix-widget-motion-handler!)
  (guarantee-procedure-of-arity handler 4 'set-fix-widget-motion-handler!)
  (vector-set!
   (fix-widget-event-handlers widget)
   (C-enum "GDK_MOTION_NOTIFY")
   (named-lambda (fix-widget-motion-handler widget GdkEvent)
     (handler widget
	      (->modifiers (C-> GdkEvent "GdkEventMotion state"))
	      (floor->exact (C-> GdkEvent "GdkEventMotion x"))
	      (floor->exact (C-> GdkEvent "GdkEventMotion y"))))))

(define ->modifiers
  (let ((names (make-vector 32 #f)))
    (define-integrable (name mask symbol)
      (vector-set! names (car (bit-mask-indices mask)) symbol))
    (name (C-enum "GDK_SHIFT_MASK")   'shift)
    (name (C-enum "GDK_LOCK_MASK")    'lock)
    (name (C-enum "GDK_CONTROL_MASK") 'control)
    (name (C-enum "GDK_MOD1_MASK")    'mod1)
    (name (C-enum "GDK_MOD2_MASK")    'mod2)
    (name (C-enum "GDK_MOD3_MASK")    'mod3)
    (name (C-enum "GDK_MOD4_MASK")    'mod4)
    (name (C-enum "GDK_MOD5_MASK")    'mod5)
    (name (C-enum "GDK_BUTTON1_MASK") 'button1)
    (name (C-enum "GDK_BUTTON2_MASK") 'button2)
    (name (C-enum "GDK_BUTTON3_MASK") 'button3)
    (name (C-enum "GDK_BUTTON4_MASK") 'button4)
    (name (C-enum "GDK_BUTTON5_MASK") 'button5)
    (name (C-enum "GDK_SUPER_MASK")   'super)
    (name (C-enum "GDK_HYPER_MASK")   'hyper)
    (name (C-enum "GDK_META_MASK")    'meta)
    (name (C-enum "GDK_RELEASE_MASK") 'release)
    (named-lambda (->modifiers num)
      (map! (lambda (i) (vector-ref names i)) (bit-mask-indices num)))))

(define (set-fix-widget-button-handler! widget type handler)
  (guarantee-fix-widget widget 'set-fix-widget-button-handler!)
  (guarantee-procedure-of-arity handler 6 'set-fix-widget-button-handler!)
  (let ((index (->button-event-type type 'set-fix-widget-button-handler!))
	(handler (make-button-handler handler)))
    (vector-set! (fix-widget-event-handlers widget) index handler)))

(define (make-button-handler handler)
  (named-lambda (fix-widget-button-handler widget GdkEvent)
    (handler widget
	     (button-event-type->name (C-> GdkEvent "GdkEvent any type"))
	     (C-> GdkEvent "GdkEventButton button")
	     (->modifiers (C-> GdkEvent "GdkEventButton state"))
	     (floor->exact (C-> GdkEvent "GdkEventButton x"))
	     (floor->exact (C-> GdkEvent "GdkEventButton y")))))

(define (->button-event-type type operator)
  (guarantee symbol? type operator)
  (case type
    ((PRESS) (C-enum "GDK_BUTTON_PRESS"))
    ((RELEASE) (C-enum "GDK_BUTTON_RELEASE"))
    ((DOUBLE-PRESS) (C-enum "GDK_2BUTTON_PRESS"))
    ((TRIPLE-PRESS) (C-enum "GDK_3BUTTON_PRESS"))
    (else (error:wrong-type-argument
	   type "a button event type (press, release, double-press or triple-press)"
	   operator))))

(define (button-event-type->name type)
  (cond ((= type (C-enum "GDK_BUTTON_PRESS")) 'PRESS)
	((= type (C-enum "GDK_BUTTON_RELEASE")) 'RELEASE)
	((= type (C-enum "GDK_2BUTTON_PRESS")) 'DOUBLE-PRESS)
	((= type (C-enum "GDK_3BUTTON_PRESS")) 'TRIPLE-PRESS)
	(else 'BOGUS)))

(define-class (<fix-layout> (constructor () (width height bgcolor)))
    (<fix-widget>)

  ;; Scrollbar widgets.
  (vadjustment define standard initial-value #f)
  (hadjustment define standard initial-value #f)
  (scroll-step define accessor initializer (lambda () (cons 10 20)))

  ;; Scrollable extent (drawing size), in logical device coords.
  (scrollable-extent define accessor
		     initializer (lambda () (make-fix-rect 0 0 100 100)))

  ;; Scroll offset (and window size) in logical device coordinates.
  ;; (The size should match the window geometry.)
  (view define accessor initializer (lambda () (make-fix-rect 0 0)))

  (drawing define standard
	   modifier %set-fix-layout-drawing!
	   initial-value #f))

(define-guarantee fix-layout "a <fix-layout>")

(define-method initialize-instance ((widget <fix-layout>) width height bgcolor)
  (call-next-method widget width height bgcolor)
  (%trace "; (initialize-instance <fix-layout>) "widget" "width"x"height"\n")
  (assert-glib-locked '(initialize-instance <fix-layout>))
  (set-fix-rect! (fix-layout-view widget) 0 0 width height)
  (set-gtk-widget-draw-callback! widget layout-draw-callback)
  (set-scm-widget-set-scroll-adjustments-callback! widget adjustments-callback)
  (C-call "gtk_widget_set_can_focus" (gobject-alien widget) 1)
  widget)

(define-method gtk-widget-destroy-callback ((layout <fix-layout>))
  (call-next-method layout)
  (let ((drawing (fix-layout-drawing layout)))
    (if drawing (fix-drawing-remove-widget! drawing layout))))

(define-integrable (fix-layout-hstep-incr widget)
  (car (fix-layout-scroll-step widget)))

(define-integrable (fix-layout-vstep-incr widget)
  (cdr (fix-layout-scroll-step widget)))

(define (fix-layout-clip-area layout cairo)
  ;; The cairo context is clipped to the exposed area in widget
  ;; coords (window coordinates).
  (cairo-clip-extents
   cairo
   (lambda (x1. y1. x2. y2.)
     (let ((x1 (floor->exact x1.))
	   (y1 (floor->exact y1.))
	   (x2 (floor->exact x2.))
	   (y2 (floor->exact y2.))
	   (view (fix-layout-view layout)))
       (make-fix-rect (fix:+ x1 (fix-rect-x view))
		      (fix:+ y1 (fix-rect-y view))
		      (fix:- x2 x1)
		      (fix:- y2 y1))))))

(define (layout-draw-callback layout cr)
  (set-alien/ctype! cr '|cairo_t|)
  (%trace2 ";draw "layout" at "
	   (cairo-clip-extents
	    cr (lambda (min-x min-y max-x max-y)
		 (define-integrable n->s number->string)
		 (string-append (n->s min-x)","(n->s min-y)
				" "(n->s (- max-x min-x))
				"x"(n->s (- max-y min-y)))))
	   "\n")
  (let ((window (fix-widget-window layout))
	(drawing (fix-layout-drawing layout))
	(area (fix-layout-clip-area layout cr)))
    (%trace2 ";  view: "(fix-rect-string (fix-layout-view layout))"\n")
    (%trace2 ";  area: "(fix-rect-string area)"\n")
    (let ((bg (fix-widget-%background-color layout)))
      (if (color? bg)
	  (begin
	    (cairo-save cr)
	    (cairo-set-source-color cr bg)
	    (cairo-paint cr)
	    (cairo-restore cr))))
    (if drawing
	(for-each
	  (lambda (ink)
	    (if (fix-ink-in? ink layout area)
		(begin
		  (cairo-save cr)
		  (fix-ink-draw-callback ink layout window cr area)
		  (cairo-restore cr))))
	  (fix-drawing-display-list drawing))
	(%trace2 ";  no drawing\n"))))

(define (set-fix-layout-scroll-size! widget width height)
  ;; Tells WIDGET to adjust its scrollable extent.  Notifies any
  ;; scrollbars.
  (guarantee-fix-layout widget 'set-fix-layout-scroll-size!)
  (guarantee-size width 'set-fix-layout-scroll-size!)
  (guarantee-size height 'set-fix-layout-scroll-size!)
  (let ((extent (fix-layout-scrollable-extent widget)))
    (if (not (and (fix:= width (fix-rect-width extent))
		  (fix:= height (fix-rect-height extent))))
	(begin
	  (set-fix-rect-size! extent width height)
	  (if (fix-widget-realized? widget)
	      (adjust-adjustments widget))))))

(define (fix-layout-scroll-to! widget x y)
  (guarantee-fix-layout widget 'fix-layout-scroll-to!)
  (guarantee fixnum? x 'fix-layout-scroll-to!)
  (guarantee fixnum? y 'fix-layout-scroll-to!)
  (scroll widget x y))

(define (fix-layout-scroll-nw! widget extent)
  (let ((view (fix-layout-view widget)))
    (cond ((not (fix-rect-nominal? extent))
	   (error "Undefined extent:" extent))
	  ((not (fix-rect-nominal? view))
	   (error "Undefined extent:" view))
	  ((fix-rect-contains? view extent)
	   unspecific)
	  (else
	   (with-fix-rect-bounds
	    view
	    (lambda (min-x-view max-x-view min-y-view max-y-view)
	      (with-fix-rect-bounds
	       extent
	       (lambda (min-x-extent max-x-extent min-y-extent max-y-extent)
		 (let ((delta-y (cond ((fix:< min-y-extent min-y-view)
				       ;; extent is too low (N), go low
				       (fix:- min-y-extent min-y-view))
				      ((fix:< max-y-view max-y-extent)
				       ;; extent is too high (S), go high
				       (fix:- max-y-extent max-y-view))
				      (else 0)))
		       (delta-x (cond ((fix:< min-x-extent min-x-view)
				       ;; extent is too low (W), go low
				       (fix:- min-x-extent min-x-view))
				      ((fix:< max-x-view max-x-extent)
				       ;; extent is too high (E), go high
				       (fix:- max-x-extent max-x-view))
				      (else 0))))
		   (fix-layout-scroll-to!
		    widget
		    (fix:+ min-x-view delta-x)
		    (fix:+ min-y-view delta-y)))))))))))

(define (scroll widget new-x new-y)
  ;; Scroll if more than 25% will remain in the window, else jump.
  (assert-glib-locked 'scroll)
  (if (fix-widget-realized? widget)
      (let ((view (fix-layout-view widget)))
	(let ((old-x (fix-rect-x view))
	      (old-y (fix-rect-y view)))
	  (let ((dx (fix:- new-x old-x))
		(dy (fix:- new-y old-y)))
	    (if (not (and (fix:zero? dx) (fix:zero? dy)))
		(let ((width (fix-rect-width view))
		      (height (fix-rect-height view))
		      (gdkwindow (fix-widget-window widget)))
		  (let ((remaining-width (fix:- width (fix:abs dy)))
			(remaining-height (fix:- height (fix:abs dx))))
		    (if (or (fix:negative? remaining-width)
			    (fix:negative? remaining-height)
			    (< 0.25 (/ (fix:* remaining-width remaining-height)
				       (fix:* width           height))))
			(C-call "gdk_window_scroll"
				gdkwindow (fix:negate dx) (fix:negate dy))
			(C-call "gtk_widget_queue_draw"
				(gobject-alien widget)))
		    (set-fix-rect-position! view new-x new-y)
		    (adjust-adjustments widget))
		  ;;(C-call "gdk_window_process_updates" gdkwindow 0)
		  )))))))

(define (set-fix-layout-scroll-step! widget width height)
  (guarantee-fix-layout widget 'set-fix-layout-scroll-step!)
  (guarantee positive-fixnum? width 'set-fix-layout-scroll-step!)
  (guarantee positive-fixnum? height 'set-fix-layout-scroll-step!)
  (let ((width.height (fix-layout-scroll-step widget)))
    (set-car! width.height width)
    (set-cdr! width.height height))
  (if (fix-widget-realized? widget)
      (adjust-adjustments widget)))

(define (set-fix-layout-drawing! widget drawing x y)
  (assert-glib-locked 'set-fix-layout-drawing!)
  (guarantee-fix-layout widget 'set-fix-layout-drawing!)
  (guarantee-fix-drawing drawing 'set-fix-layout-drawing!)
  (guarantee fixnum? x 'set-fix-layout-drawing!)
  (guarantee fixnum? y 'set-fix-layout-drawing!)
  (let* ((old (fix-layout-drawing widget))
	 (view (fix-layout-view widget)))
    (if (and (eq? drawing old) (fix-rect-at-point? view x y))
	unspecific
	(let ((extent (fix-drawing-extent drawing))
	      (scrollable (fix-layout-scrollable-extent widget)))
	  (set-fix-rect-position! view x y)
	  (set-fix-rect-size! scrollable
			      (fix-rect-width extent) (fix-rect-height extent))
	  (if old (fix-drawing-remove-widget! old widget))
	  (if drawing (fix-drawing-add-widget! drawing widget))
	  (%set-fix-layout-drawing! widget drawing)
	  (if (fix-widget-realized? widget)
	      (begin
		(adjust-adjustments widget)
		(C-call "gtk_widget_queue_draw" (gobject-alien widget))))))))

;;; Callbacks.

(define-method fix-widget-new-geometry-callback ((widget <fix-layout>))
  (call-next-method widget)
  (%trace "; (fix-widget-new-geometry-callback <fix-layout>) "widget"\n")
  (let ((geom (fix-widget-geometry widget)))
    (set-fix-rect-size! (fix-layout-view widget)
			(fix-rect-width geom) (fix-rect-height geom)))
  (adjust-adjustments widget))

(define-method fix-widget-realize-callback ((widget <fix-layout>))
  (call-next-method widget)
  (%trace "; (fix-widget-realize-callback <fix-layout>) "widget"\n")
  (assert-glib-locked '(fix-widget-realize-callback <fix-layout>))
  #;(let ((style (gtk-widget-style-context widget)))
    (C-call "gtk_style_context_set_background"
	    style (fix-widget-window widget)))
  (adjust-adjustments widget))

(define (adjustments-callback widget hGtkAdjustment vGtkAdjustment)
  (%trace2 ";set-scroll-adjustments "widget
	   " "hGtkAdjustment" "vGtkAdjustment"\n")
  (%trace ";adjustments:"
	  " 0x"(alien/address-string hGtkAdjustment)
	  " 0x"(alien/address-string vGtkAdjustment)"\n")
  (connect-adjustment (fix-layout-hadjustment widget) hGtkAdjustment
		      widget set-fix-layout-hadjustment!)
  (connect-adjustment (fix-layout-vadjustment widget) vGtkAdjustment
		      widget set-fix-layout-vadjustment!)
  (if (fix-widget-realized? widget)
      (adjust-adjustments widget)))

(define (connect-adjustment old-adjustment new-alien widget setter)
  (assert-glib-locked 'connect-adjustment)
  (let ((old-alien (and old-adjustment (gobject-alien old-adjustment))))
    ;; Disconnect.
    (cond ((not old-adjustment))
	  ((alien=? new-alien old-alien))
	  (else
	   (gobject-unref! old-adjustment)))
    ;; Connect.
    (cond ((alien-null? new-alien))
	  ((and old-alien (alien=? new-alien old-alien)))
	  (else
	   (let ((new-adjustment (make-gtk-adjustment)))
	     (copy-alien-address! (gobject-alien new-adjustment) new-alien)
	     (C-call "g_object_ref_sink" new-alien new-alien)
	     (setter widget new-adjustment)
	     (g-signal-connect
	      new-adjustment (C-callback "value_changed")
	      (make-adjustment-value-changed-callback widget)))))))

(define (make-adjustment-value-changed-callback widget)
  (named-lambda (fix-layout-adjustment-value-changed-callback adjustment)
    (%trace2 ";adjustment-value-changed "widget" "adjustment"\n")
    (assert-glib-locked 'make-adjustment-value-changed-callback)
    (let ((window-extent (fix-layout-view widget))
	  (vadjustment (fix-layout-vadjustment widget))
	  (hadjustment (fix-layout-hadjustment widget))
	  (value (floor->exact
		  (C-call "gtk_adjustment_get_value"
			  (gobject-alien adjustment)))))
      (cond ((eq? adjustment vadjustment)
	     (%trace2 ";vadjustment to "value"\n")
	     (scroll widget (fix-rect-x window-extent) value))
	    ((eq? adjustment hadjustment)
	     (%trace2 ";hadjustment to "value"\n")
	     (scroll widget value (fix-rect-y window-extent)))
	    (else (warn "Unexpected adjustment:" adjustment))))))

(define (adjust-adjustments widget)
  ;; Called after the widget gets new adjustment(s) or its size or
  ;; scrollable extent changes.

  (%trace2 ";adjust horizontal adjustment "widget"\n")
  (adjust-adjustment widget (fix-layout-hadjustment widget)
		     fix-rect-width fix-rect-x
		     (lambda (rect x)
		       (set-fix-rect-x! rect x)
		       (gtk-widget-queue-draw widget))
		     fix-layout-hstep-incr)
  (%trace2 ";adjust vertical adjustment "widget"\n")
  (adjust-adjustment widget (fix-layout-vadjustment widget)
		     fix-rect-height fix-rect-y
		     (lambda (rect y)
		       (set-fix-rect-y! rect y)
		       (gtk-widget-queue-draw widget))
		     fix-layout-vstep-incr))

(define (adjust-adjustment widget adj
			   fix-rect-size fix-rect-low set-fix-rect-low!
			   widget-step-incr)
  (if (and adj (gobject-live? adj))
      (let ((view (fix-layout-view widget))
	    (extent (fix-layout-scrollable-extent widget)))
	(let ((view-size (fix-rect-size view))
	      (extent-size (fix-rect-size extent))
	      (extent-low (fix-rect-low extent))
	      (step-incr (widget-step-incr widget)))
	  (fix-layout-adjustment-parameters
	   widget view-size extent-size extent-low
	   (lambda (low page-size)

	     (define-integrable (clamped-value! low high)
	       (let ((value (fix-rect-low view)))
		 (cond ((fix:< value low)
			(set-fix-rect-low! view low)
			low)
		       ((fix:< high value)
			(set-fix-rect-low! view high)
			high)
		       (else value))))

	     (let ((high (fix:+ extent-low extent-size))
		   (page-incr (fix:max 1 (fix:- page-size step-incr))))
		(let ((value (clamped-value! low (fix:- high page-size))))
		  (%trace2 ";  adjustment: "low" "value" "high" "page-size"\n")
		  (set-gtk-adjustment! adj value low high
				       page-size step-incr page-incr)))))))))

(define-generic fix-layout-adjustment-parameters (widget
						  view-size extent-size
						  extent-low receiver))

(define-method fix-layout-adjustment-parameters ((widget <fix-layout>)
						 view-size extent-size
						 extent-low receiver)
  (if (fix:< view-size extent-size)
      ;; Drawing is larger than viewport: thumb (page) is viewport.
      (let ((low extent-low)
	    (page-size view-size))
	(%trace2 ";  large-drawing:"(fix-layout-scrollable-extent widget)
		 " view:"(fix-layout-view widget)"\n")
	(receiver low page-size))
      ;; Viewport is larger than drawing: thumb (page) is drawing.
      (let ((low (fix:- extent-low (fix:- view-size extent-size)))
	    (page-size extent-size))
	(%trace2 ";  drawing:"(fix-layout-scrollable-extent widget)
		 " large-view:"(fix-layout-view widget)"\n")
	(receiver low page-size))))

(define-class (<fix-drawing> (constructor () no-init))
    ()
  (extent define accessor initializer (lambda () (make-fix-rect 0 0 0 0)))
  (widgets define standard initial-value '())
  (display-list define standard initial-value '()))

(define-guarantee fix-drawing "a <fix-drawing>")

(define (drawing-damage ink #!optional rect)
  ;; Invalidates any widget extents affected by RECT in INK.  By
  ;; default, RECT is INK's entire extent.
  (assert-glib-locked 'drawing-damage)
  (let ((extent (if (default-object? rect)
		  (fix-ink-extent ink)
		  rect))
	(drawing (fix-ink-drawing ink)))
    (%trace2 ";drawing-damage "ink" "(fix-rect-string extent)"\n")

    (cond ((not drawing))
	  ((not (fix-rect-nominal? extent))
	   (error "Cannot damage ill-defined extent:" ink))
	  ((and (not (fix:zero? (fix-rect-width extent)))
		(not (fix:zero? (fix-rect-height extent))))
	   (for-each
	     (lambda (widget)
	       (if (not (gtk-widget-destroyed? widget))
		   (let ((intersect (let ((v (fix-layout-view widget)))
				      (and (fix-rect-nominal? v)
					   (window-intersection v extent)))))
		     (if intersect
			 (C-call "gtk_widget_queue_draw_area"
				 (gobject-alien widget)
				 (fix-rect-x intersect) (fix-rect-y intersect)
				 (fix-rect-width intersect)
				 (fix-rect-height intersect))))))
	     (let ((widgets (fix-ink-widgets ink)))
	       (if (eq? #t widgets)
		   (fix-drawing-widgets drawing)
		   widgets)))))))

(define (fix-drawing-pick-list drawing widget x y)
  (filter (lambda (ink)
	    (and (fix-ink-in-widget? ink widget)
		 (point-in-fix-rect? x y (fix-ink-extent ink))))
	  (fix-drawing-display-list drawing)))

(define (fix-ink-in? ink widget area)
  (declare (integrate-operator fix-ink-in?))
  (and (fix-ink-in-widget? ink widget)
       (let ((extent (fix-ink-extent ink)))
	 (and (fix-rect-nominal? extent)
	      (fix-rect-intersect? extent area)))))

(define-integrable (fix-ink-in-widget? ink widget)
  (let ((widgets (fix-ink-widgets ink)))
    (or (eq? #t widgets)
	(memq widget widgets))))

(define-generic fix-ink-draw-callback (ink widget window cr exposed-area)
  ;; Due to the checks in layout-draw-callback, methods of this generic can
  ;; assume expose-area and the ink's extent are intersecting, and INK
  ;; is visible in the WIDGET.  Methods may also assume the widget is
  ;; realized and its window's cairo's clipping is already set.  The
  ;; widget's scroll offset (view extent) is also set.
  )

(define (fix-drawing-add-widget! drawing widget)
  (guarantee-fix-drawing drawing 'fix-drawing-add-widget!)
  (guarantee-fix-layout widget 'fix-drawing-add-widget!)
  (let ((widgets (fix-drawing-widgets drawing)))
    (if (not (memq widget widgets))
	(set-fix-drawing-widgets! drawing (cons widget widgets)))))

(define (fix-drawing-remove-widget! drawing widget)
  (guarantee-fix-drawing drawing 'fix-drawing-remove-widget!)
  (guarantee-fix-layout widget 'fix-drawing-remove-widget!)
  (set-fix-drawing-widgets! drawing (delq! widget
					   (fix-drawing-widgets drawing))))

(define (fix-drawing-add-ink! drawing ink #!optional where)
  (guarantee-fix-drawing drawing 'fix-drawing-add-ink!)
  (guarantee-fix-ink ink 'fix-drawing-add-ink!)

  (let ((where (and (not (default-object? where)) where)))
    (if (fix-ink-drawing ink) (error "Already in a drawing:" ink))
    (cond ((or (eq? #f where)
	       (eq? 'TOP where))
	   (set-fix-drawing-display-list!
	    drawing (append! (fix-drawing-display-list drawing) (list ink))))
	  ((eq? 'BOTTOM where)
	   (set-fix-drawing-display-list!
	    drawing (cons ink (fix-drawing-display-list drawing))))
	  ((fix-ink? where)
	   (let loop ((inks (fix-drawing-display-list drawing))
		      (prev #f))
	     (if (null? inks)
		 (error "Ink not found in drawing:" ink drawing)
		 (let ((i (car inks)))
		   (if (eq? where i)
		       (if (pair? prev)
			   (set-cdr! prev (cons ink inks))
			   (set-fix-drawing-display-list! drawing
							  (cons ink inks)))
		       (loop (cdr inks) inks))))))
	  (else (error:wrong-type-argument
		 where
		 (string-append "display list location"
				", one of: #F, TOP, BOTTOM, or an <ink>"
				" already in the drawing's display list")
		 'fix-drawing-add-ink!))))
  (set-fix-ink-drawing! ink drawing)
  (drawing-damage ink))

(define (set-fix-drawing-size! drawing width height)
  (guarantee-fix-drawing drawing 'set-fix-drawing-size!)
  (guarantee-size width 'set-fix-drawing-size!)
  (guarantee-size height 'set-fix-drawing-size!)
  (set-fix-rect-size! (fix-drawing-extent drawing) width height)
  (for-each
    (lambda (widget) (set-fix-layout-scroll-size! widget width height))
    (fix-drawing-widgets drawing)))

(define (set-drawing-extent! drawing rectangle)
  (let ((extent (fix-drawing-extent drawing)))
    (set-fix-rect! extent
		   (fix-rect-x rectangle)
		   (fix-rect-y rectangle)
		   (fix-rect-width rectangle)
		   (fix-rect-height rectangle))))

(define-class <fix-ink>
    ()
  (extent define standard initializer (lambda () (make-fix-rect 0 0 0 0)))
  (drawing define standard initial-value #f)

  ;; A list of widgets in which the ink should be drawn.  #t if the
  ;; ink should be visible in all views of the drawing.
  (widgets define standard initial-value #t
	   modifier set-fix-ink-%widgets!))

(define-guarantee fix-ink "a <fix-ink>")

(define (set-fix-ink-%position! ink x y)
  (assert-glib-locked 'set-fix-ink-%position!)
  (let ((extent (fix-ink-extent ink)))
    (without-interruption
     (lambda ()
       (if (not (and (fix:= x (fix-rect-x extent))
		     (fix:= y (fix-rect-y extent))))
	   (begin
	     (drawing-damage ink)
	     (set-fix-rect-position! extent x y)
	     (drawing-damage ink)))))))

(define (set-fix-ink! ink x y width height)
  (assert-glib-locked 'set-fix-ink!)
  (let ((extent (fix-ink-extent ink)))
    (without-interruption
     (lambda ()
       (if (not (and (fix:= x (fix-rect-x extent))
		     (fix:= y (fix-rect-y extent))
		     (fix:= width (fix-rect-width extent))
		     (fix:= height (fix-rect-height extent))))
	   (begin
	     (drawing-damage ink)
	     (set-fix-rect! extent x y width height)
	     (drawing-damage ink)))))))

(define (set-fix-ink-widgets! ink widgets)
  (assert-glib-locked 'set-fix-ink-widgets!)
  (without-interruption
   (lambda ()
     (if (not (equal? widgets (fix-ink-widgets ink)))
	 (begin
	   (drawing-damage ink)
	   (set-fix-ink-%widgets! ink widgets)
	   (drawing-damage ink))))))

(define (fix-ink-remove! ink)
  (assert-glib-locked 'fix-ink-remove!)
  (guarantee-fix-ink ink 'fix-ink-remove!)
  (let ((drawing (fix-ink-drawing ink)))
    (cond ((not drawing) unspecific)
	  ((not (memq ink (fix-drawing-display-list drawing)))
	   (warn "Could not remove ink:" ink drawing))
	  (else
	   (without-interruption
	    (lambda ()
	      (set-fix-drawing-display-list!
	       drawing (delq! ink (fix-drawing-display-list drawing)))
	      (drawing-damage ink)
	      (set-fix-ink-drawing! ink #f)))))))

;; For the convenience of SWAT's canvas item group, mostly.
(define-generic fix-ink-move! (ink dx dy))

;; Use this to define your fix-ink-move! method iff your extent is all
;; that needs to be updated when you move.  This is not a default
;; method, else things might (only) appear to work.
(define-integrable (generic-fix-ink-move! ink dx dy)
  (assert-glib-locked 'generic-fix-ink-move!)
  (without-interruption
   (lambda ()
     (let ((extent (fix-ink-extent ink)))
       (drawing-damage ink)
       (fix-rect-move! extent dx dy)
       (drawing-damage ink)))))

;; This kind of ink draws (outlines and/or fills) a shape according to
;; a variety of options.  These can be converted LATE, e.g. on expose,
;; when there is a realized widget.

(define-class <draw-ink>
    (<fix-ink>)
  ;; Alist of option names (symbols) X values (whatnot).
  (options define standard initial-value '()))

(define-guarantee draw-ink "a <draw-ink>")

;;; For draw-ink expose handlers (without-interrupts in gtk-thread).

(define-integrable (get-option ink name default)
  (let ((entry (assq name (draw-ink-options ink))))
    (if entry (cdr entry) default)))

(define-integrable (set-option!? ink name value)
  ;; If VALUE is null, the option is unset.  If the option value is
  ;; already VALUE, returns #f (else #t).
  (let* ((options (draw-ink-options ink))
	 (entry (assq name options))
	 (old-value (if entry (cdr entry) '())))
    (if (equal? value old-value)
	#f
	(begin
	  (if entry
	      (if (null? value)
		  (set-draw-ink-options! ink (delq! entry options))
		  (set-cdr! entry value))
	      (set-draw-ink-options! ink (cons (cons name value) options)))
	  #t))))

(define-class (<line-ink> (constructor ()))
    (<draw-ink>)
  (vector define standard initializer (lambda () (make-fix-rect 0 0 0 0))))

(define-guarantee line-ink "a <line-ink>")

(define-method fix-ink-draw-callback ((ink <line-ink>) widget window cr area)
  (declare (ignore window area))
  (%trace2 ";drawing "ink" on "widget"\n")
  (let ((view (fix-layout-view widget))
	(vector (line-ink-vector ink)))
    (with-fix-rect
     vector
     (lambda (x y dx dy)
       (let ((x (fix:- x (fix-rect-x view)))
	     (y (fix:- y (fix-rect-y view))))
	 (cairo-move-to cr x y)
	 (cairo-rel-line-to cr dx dy))))
    (set-line-options! cr ink)
    (let ((color (get-option ink 'DASH-COLOR '())))
      (if (not (null? color))
	  (begin
	    (cairo-save cr)
	    (cairo-set-source-color cr color)
	    (cairo-stroke-preserve cr)
	    (cairo-restore cr))))
    (set-line-dashes! cr ink)
    (cairo-stroke cr)))

(define (set-line-options! cr ink)
  (for-each
    (lambda (entry)
      (let ((name (car entry))
	    (value (cdr entry)))
	(case name
	  ((COLOR) (cairo-set-source-color cr value))
	  ;;((LINE-CAP) ...)
	  ;;((LINE-JOIN) ...)
	  ;;((LINE-MITER-LIMIT) ...)
	  ((LINE-WIDTH) (cairo-set-line-width cr value)))))
    (draw-ink-options ink)))

(define (set-line-dashes! cr ink)
  (let ((entry (assq 'DASHES (draw-ink-options ink))))
    (if entry
	(cairo-set-dash cr (cdr entry)))))

(define-integrable (half-line-width ink)
  (fix:max 1 (fix:1+ (floor->exact (quotient (get-option ink 'LINE-WIDTH 1.)
					     2)))))

(define (recache-line-extent! ink)
  (assert-without-interruption 'recache-line-extent!)
  (assert-glib-locked 'recache-line-extent!)
  (with-fix-rect
   (line-ink-vector ink)
   (lambda (x1 y1 dx dy)
     (let ((x2 (fix:+ x1 dx))
	   (y2 (fix:+ y1 dy)))
       (let ((min-x (fix:min x1 x2))
	     (min-y (fix:min y1 y2))
	     (max-x (fix:max x1 x2))
	     (max-y (fix:max y1 y2))
	     (lw/2 (half-line-width ink)))
	 (drawing-damage ink)
	 (set-fix-rect-bounds! (fix-ink-extent ink)
			   (fix:- min-x lw/2)
			   (fix:+ max-x lw/2)
			   (fix:- min-y lw/2)
			   (fix:+ max-y lw/2))
	 (drawing-damage ink))))))

(define (set-line-ink! ink x1 y1 x2 y2)
  (assert-glib-locked 'set-line-ink!)
  (guarantee fixnum? x1 'set-line-ink!)
  (guarantee fixnum? y1 'set-line-ink!)
  (guarantee fixnum? x2 'set-line-ink!)
  (guarantee fixnum? y2 'set-line-ink!)
  (without-interruption
   (lambda ()
     (let ((vector (line-ink-vector ink))
	   (dx (fix:- x2 x1))
	   (dy (fix:- y2 y1)))
       (if (not (and (fix:= x1 (fix-rect-x vector))
		     (fix:= y1 (fix-rect-y vector))
		     (fix:= dx (fix-rect-width vector))
		     (fix:= dy (fix-rect-height vector))))
	   (begin
	     (set-fix-rect! vector x1 y1 dx dy)
	     (recache-line-extent! ink)))))))

(define-method fix-ink-move! ((ink <line-ink>) dx dy)
  (assert-glib-locked '(fix-ink-move! <line-ink>))
  (without-interruption
   (lambda ()
     (let ((vector (line-ink-vector ink))
	   (extent (fix-ink-extent ink)))
       (drawing-damage ink)
       (fix-rect-move! vector dx dy)
       (fix-rect-move! extent dx dy)
       (drawing-damage ink)))))

(define (line-ink-width ink)
  (guarantee-line-ink ink 'line-ink-width)
  (get-option ink 'LINE-WIDTH '()))

(define (set-line-ink-width! ink width)
  (assert-glib-locked 'set-line-ink-width!)
  (guarantee-line-ink ink 'set-line-ink-width!)
  (guarantee positive-fixnum? width 'set-line-ink-width!)
  (without-interruption
   (lambda ()
     (if (set-option!? ink 'LINE-WIDTH (->flonum width))
	 (recache-line-extent! ink)))))

(define (line-ink-color ink)
  (guarantee-line-ink ink 'line-ink-color)
  (get-option ink 'COLOR '()))

(define (set-line-ink-color! ink color)
  (assert-glib-locked 'set-line-ink-color!)
  (guarantee-line-ink ink 'set-line-ink-color!)
  (let ((color (->color color 'set-line-ink-color!)))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'COLOR color)
	   (drawing-damage ink))))))

(define (line-ink-dash-color ink)
  (guarantee-line-ink ink 'line-ink-dash-color)
  (get-option ink 'DASH-COLOR '()))

(define (set-line-ink-dash-color! ink color)
  (assert-glib-locked 'set-line-ink-dash-color!)
  (guarantee-line-ink ink 'set-line-ink-dash-color!)
  (let ((color (cond ((eq? color '()) '())
		     (else (->color color 'set-line-ink-dash-color!)))))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'DASH-COLOR color)
	   (drawing-damage ink))))))

(define (line-ink-dashes ink)
  (guarantee-line-ink ink 'line-ink-dash-color)
  (get-option ink 'DASHES '()))

(define (set-line-ink-dashes! ink lengths)
  (assert-glib-locked 'set-line-ink-dashes!)
  (guarantee-line-ink ink 'set-line-ink-dashes!)
  (guarantee-list-of-type lengths flo:flonum?
			  "list of flonums" 'set-line-ink-dashes!)
  (without-interruption
   (lambda ()
     (if (set-option!? ink 'DASHES lengths)
	 (drawing-damage ink)))))

(define-class (<rectangle-ink> (constructor ()))
    (<draw-ink>)
  (rect define standard initializer (lambda () (make-fix-rect 0 0 0 0))))

(define-guarantee rectangle-ink "a <rectangle-ink>")

(define-method fix-ink-draw-callback ((ink <rectangle-ink>)
				      widget window cr area)
  (declare (ignore window area))
  (%trace2 ";drawing "ink" on "widget"\n")
  (let ((view (fix-layout-view widget))
	(rect (rectangle-ink-rect ink)))
    (with-fix-rect
     rect
     (lambda (x y width height)
       (let ((x (fix:- x (fix-rect-x view)))
	     (y (fix:- y (fix-rect-y view))))
	 (cairo-rectangle cr x y width height)
	 (let ((fill (get-option ink 'FILL '())))
	   (if (not (null? fill))
	       (begin
		 (cairo-save cr)
		 (set-fill-options! cr ink)
		 (cairo-fill-preserve cr)
		 (cairo-restore cr))))
	 (let ((outline (get-option ink 'OUTLINE '())))
	   (if (not (null? outline))
	       (begin
		 (set-outline-options! cr ink)
		 (cairo-stroke cr)))))))))

(define (set-fill-options! cr ink)
  ;; For filling ovals, rectangles...
  (for-each
    (lambda (entry)
      (let ((name (car entry))
	    (value (cdr entry)))
	(case name
	  ((FILL) (cairo-set-source-color cr value)))))
    (draw-ink-options ink)))

(define (set-outline-options! cr ink)
  (for-each
    (lambda (entry)
      (let ((name (car entry))
	    (value (cdr entry)))
	(case name
	  ((OUTLINE) (cairo-set-source-color cr value))
	  ((LINE-WIDTH) (cairo-set-line-width cr value))
	  ((DASHES) (set-line-dashes! cr ink)))))
    (draw-ink-options ink)))

(define (recache-rectangle-extent! ink)
  (with-fix-rect-bounds
   (rectangle-ink-rect ink)
   (lambda (min-x max-x min-y max-y)
     (let ((lw/2 (half-line-width ink)))
       (drawing-damage ink)
       (set-fix-rect-bounds! (fix-ink-extent ink)
			 (fix:- min-x lw/2)
			 (fix:+ max-x lw/2)
			 (fix:- min-y lw/2)
			 (fix:+ max-y lw/2))
       (drawing-damage ink)))))

(define (set-rectangle-ink! ink x y width height)
  (assert-glib-locked 'set-rectangle-ink!)
  (guarantee fixnum? x 'set-rectangle-ink!)
  (guarantee fixnum? y 'set-rectangle-ink!)
  (guarantee-size width 'set-rectangle-ink!)
  (guarantee-size height 'set-rectangle-ink!)
  (without-interruption
   (lambda ()
     (let ((rect (rectangle-ink-rect ink)))
       (if (not (and (fix:= x (fix-rect-x rect))
		     (fix:= y (fix-rect-y rect))
		     (fix:= width (fix-rect-width rect))
		     (fix:= height (fix-rect-height rect))))
	   (begin
	     (set-fix-rect! rect x y width height)
	     (recache-rectangle-extent! ink)))))))

(define (set-rectangle-ink-position! ink x y)
  (assert-glib-locked 'set-rectangle-ink-position!)
  (guarantee fixnum? x 'set-rectangle-ink-position!)
  (guarantee fixnum? y 'set-rectangle-ink-position!)
  (without-interruption
   (lambda ()
     (let ((rect (rectangle-ink-rect ink)))
       (if (not (and (fix:= x (fix-rect-x rect))
		     (fix:= y (fix-rect-y rect))))
	   (begin
	     (set-fix-rect-position! rect x y)
	     (recache-rectangle-extent! ink)))))))

(define-method fix-ink-move! ((ink <rectangle-ink>) dx dy)
  (assert-glib-locked '(fix-ink-move! <rectangle-ink>))
  (without-interruption
   (lambda ()
     (let ((rect (rectangle-ink-rect ink))
	   (extent (fix-ink-extent ink)))
       (drawing-damage ink)
       (fix-rect-move! rect dx dy)
       (fix-rect-move! extent dx dy)
       (drawing-damage ink)))))

(define (rectangle-ink-width ink)
  (guarantee-rectangle-ink ink 'rectangle-ink-width)
  (get-option ink 'LINE-WIDTH '()))

(define (set-rectangle-ink-width! ink width)
  (assert-glib-locked 'set-rectangle-ink-width!)
  (guarantee-rectangle-ink ink 'set-rectangle-ink-width!)
  (guarantee positive-fixnum? width 'set-rectangle-ink-width!)
  (without-interruption
   (lambda ()
     (if (set-option!? ink 'LINE-WIDTH (->flonum width))
	 (recache-rectangle-extent! ink)))))

(define (rectangle-ink-color ink)
  (guarantee-rectangle-ink ink 'rectangle-ink-color)
  (get-option ink 'OUTLINE '()))

(define (set-rectangle-ink-color! ink color)
  (assert-glib-locked 'set-rectangle-ink-color!)
  (guarantee-rectangle-ink ink 'set-rectangle-ink-color!)
  (let ((color (->color color 'set-rectangle-ink-color!)))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'OUTLINE color)
	   (drawing-damage ink))))))

(define (rectangle-ink-fill-color ink)
  (guarantee-rectangle-ink ink 'rectangle-ink-fill-color)
  (get-option ink 'FILL '()))

(define (set-rectangle-ink-fill-color! ink color)
  (assert-glib-locked 'set-rectangle-ink-fill-color!)
  (guarantee-rectangle-ink ink 'set-rectangle-ink-fill-color!)
  (let ((color (->color color 'set-rectangle-ink-fill-color!)))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'FILL color)
	   (drawing-damage ink))))))

(define-class (<polygon-ink> (constructor ()))
    (<draw-ink>)
  (vertices define standard initial-value '()))

(define-guarantee polygon-ink "a <polygon-ink>")

(define-method fix-ink-draw-callback ((ink <polygon-ink>) widget window cr area)
  (declare (ignore window area))
  (%trace2 ";drawing "ink" on "widget"\n")
  (let ((view (fix-layout-view widget))
	(vertices (polygon-ink-vertices ink)))
    (if (not (null? vertices))
	(let ((view-x (fix-rect-x view))
	      (view-y (fix-rect-y view)))
	  (cairo-move-to cr
			 (fix:- (caar vertices) view-x)
			 (fix:- (cdar vertices) view-y))
	  (let loop ((verts (cdr vertices)))
	    (if (not (null? verts))
		(begin
		  (cairo-line-to cr
				 (fix:- (caar verts) view-x)
				 (fix:- (cdar verts) view-y))
		  (loop (cdr verts)))))
	  (let ((fill (get-option ink 'FILL '())))
	    (if (not (null? fill))
		(begin
		  (cairo-save cr)
		  (set-fill-options! cr ink)
		  (cairo-fill-preserve cr)
		  (cairo-restore cr))))
	  (let ((outline (get-option ink 'OUTLINE '())))
	    (if (not (null? outline))
		(begin
		  (set-outline-options! cr ink)
		  (cairo-stroke cr))))))))

(define (recache-polygon-extent! ink)
  (let ((vertices (polygon-ink-vertices ink)))
    (if (null? vertices)
	(let ((extent (fix-ink-extent ink)))
	  (if (not (fix:zero? (fix-rect-width ink)))
	      (begin
		(set-fix-rect! extent 0 0 0 0)
		(drawing-damage ink))))
	(let loop ((verts vertices) (min-x #f) (max-x #f) (min-y #f) (max-y #f))
	  (if (pair? verts)
	      (let ((x (caar verts))
		    (y (cdar verts)))
		(loop (cdr verts)
		      (if min-x (fix:min min-x x) x)
		      (if max-x (fix:max max-x x) x)
		      (if min-y (fix:min min-y y) y)
		      (if max-y (fix:max max-y y) y)))
	      (let ((lw/2 (half-line-width ink)))
		(drawing-damage ink)
		(set-fix-rect-bounds! (fix-ink-extent ink)
				      (fix:- min-x lw/2)
				      (fix:+ max-x lw/2)
				      (fix:- min-y lw/2)
				      (fix:+ max-y lw/2))
		(drawing-damage ink)))))))

(define (set-polygon-ink! ink vertices)
  (assert-glib-locked 'set-polygon-ink!)
  (if (or (null? vertices)
	  (not (every (lambda (p)
			(and (pair? p) (fixnum? (car p)) (fixnum? (cdr p))))
		      vertices)))
      (error:wrong-type-argument vertices "a list of pairs of fixnums"
				 'SET-POLYGON-INK!))
  (without-interruption
   (lambda ()
     (set-polygon-ink-vertices! ink vertices)
     (recache-polygon-extent! ink))))

(define-method fix-ink-move! ((ink <polygon-ink>) dx dy)
  (assert-glib-locked '(fix-ink-move! <polygon-ink>))
  (without-interruption
   (lambda ()
     (for-each (lambda (p)
		 (set-car! p (fix:+ (car p) dx))
		 (set-cdr! p (fix:+ (cdr p) dy)))
	       (polygon-ink-vertices ink))
     (drawing-damage ink)
     (fix-rect-move! (fix-ink-extent ink) dx dy)
     (drawing-damage ink))))

(define (polygon-ink-width ink)
  (guarantee-polygon-ink ink 'polygon-ink-width)
  (get-option ink 'LINE-WIDTH '()))

(define (set-polygon-ink-width! ink width)
  (assert-glib-locked 'set-polygon-ink-width!)
  (guarantee-polygon-ink ink 'set-polygon-ink-width!)
  (guarantee positive-fixnum? width 'set-polygon-ink-width!)
  (without-interruption
   (lambda ()
     (if (set-option!? ink 'LINE-WIDTH (->flonum width))
	 (recache-polygon-extent! ink)))))

(define (polygon-ink-color ink)
  (guarantee-polygon-ink ink 'polygon-ink-color)
  (get-option ink 'OUTLINE '()))

(define (set-polygon-ink-color! ink color)
  (assert-glib-locked 'set-polygon-ink-color!)
  (guarantee-polygon-ink ink 'set-polygon-ink-color!)
  (let ((color (->color color 'set-polygon-ink-color!)))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'OUTLINE color)
	   (drawing-damage ink))))))

(define (polygon-ink-fill-color ink)
  (guarantee-polygon-ink ink 'polygon-ink-fill-color)
  (get-option ink 'FILL '()))

(define (set-polygon-ink-fill-color! ink color)
  (assert-glib-locked 'set-polygon-ink-fill-color!)
  (guarantee-polygon-ink ink 'set-polygon-ink-fill-color!)
  (let ((color (->color color 'set-polygon-ink-fill-color!)))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'FILL color)
	   (drawing-damage ink))))))


(define-integrable flo:pi (flo:* 4. (flo:atan2 1. 1.)))

(define-class (<arc-ink> (constructor ()))
    (<draw-ink>)
  (rect define standard initializer (lambda () (make-fix-rect 0 0 0 0)))
  (%start-angle define standard initial-value 0.)
  (%sweep-angle define standard initial-value (flo:* 2. flo:pi)))

(define-guarantee arc-ink "an <arc-ink>")

(define-method fix-ink-draw-callback ((ink <arc-ink>) widget window cr area)
  (declare (ignore window area))
  (%trace2 ";drawing "ink" on "widget"\n")
  (let ((view (fix-layout-view widget))
	(rect (arc-ink-rect ink)))
    (with-fix-rect
     rect
     (lambda (x y width height)
       (let ((x. (->flonum (fix:- x (fix-rect-x view))))
	     (y. (->flonum (fix:- y (fix-rect-y view))))
	     (width. (->flonum width))
	     (height. (->flonum height))
	     (start. (arc-ink-%start-angle ink))
	     (end. (flo:+ (arc-ink-%start-angle ink)
			  (arc-ink-%sweep-angle ink))))
	 (cairo-save cr)
	 (cairo-translate cr
			  (flo:+ x. (flo:/ width. 2.))
			  (flo:+ y. (flo:/ height. 2.)))
	 (cairo-scale cr (flo:/ width. 2.) (flo:/ height. 2.))
	 (cairo-arc cr 0. 0. 1. start. end.)
	 (cairo-restore cr)
	 (let ((fill (get-option ink 'FILL '())))
	   (if (not (null? fill))
	       (begin
		 (set-fill-options! cr ink)
		 (cairo-fill-preserve cr))))
	 (let ((outline (get-option ink 'OUTLINE '())))
	   (if (not (null? outline))
	       (begin
		 (set-outline-options! cr ink)
		 (cairo-stroke cr)))))))))

(define (recache-arc-extent! ink)
  (with-fix-rect-bounds
   (arc-ink-rect ink)
   (lambda (min-x max-x min-y max-y)
     (let ((lw/2 (half-line-width ink)))
       (drawing-damage ink)
       (set-fix-rect-bounds! (fix-ink-extent ink)
			 (fix:- min-x lw/2)
			 (fix:+ max-x lw/2)
			 (fix:- min-y lw/2)
			 (fix:+ max-y lw/2))
       (drawing-damage ink)))))

(define (set-arc-ink! ink x y width height)
  (assert-glib-locked 'set-arc-ink!)
  (guarantee fixnum? x 'set-arc-ink!)
  (guarantee fixnum? y 'set-arc-ink!)
  (guarantee-size width 'set-arc-ink!)
  (guarantee-size height 'set-arc-ink!)
  (without-interruption
   (lambda ()
     (let ((rect (arc-ink-rect ink)))
       (if (not (and (fix:= x (fix-rect-x rect))
		     (fix:= y (fix-rect-y rect))
		     (fix:= width (fix-rect-width rect))
		     (fix:= height (fix-rect-height rect))))
	   (begin
	     (set-fix-rect! rect x y width height)
	     (recache-arc-extent! ink)))))))

(define-method fix-ink-move! ((ink <arc-ink>) dx dy)
  (assert-glib-locked '(fix-ink-move! <arc-ink>))
  (without-interruption
   (lambda ()
     (let ((rect (arc-ink-rect ink))
	   (extent (fix-ink-extent ink)))
       (drawing-damage ink)
       (fix-rect-move! rect dx dy)
       (fix-rect-move! extent dx dy)
       (drawing-damage ink)))))

(define (arc-ink-start-angle arc)
  (guarantee-arc-ink arc 'arc-ink-start-angle)
  (flo:* (flo:/ (arc-ink-%start-angle arc) flo:pi) 180.))

(define (set-arc-ink-start-angle! arc degrees)
  (guarantee-arc-ink arc 'set-arc-ink-start-angle!)
  (guarantee real? degrees 'set-arc-ink-start-angle!)
  (let ((new (flo:* (->flonum degrees) (flo:/ flo:pi 180.))))
    (if (not (flo:= new (arc-ink-%start-angle arc)))
	(begin
	  (set-arc-ink-%start-angle! arc new)
	  (drawing-damage arc)))))

(define (arc-ink-sweep-angle arc)
  (guarantee-arc-ink arc 'arc-ink-sweep-angle)
  (flo:* (flo:/ (arc-ink-%sweep-angle arc) flo:pi) 180.))

(define (set-arc-ink-sweep-angle! arc degrees)
  (guarantee-arc-ink arc 'set-arc-ink-sweep-angle!)
  (guarantee real? degrees 'set-arc-ink-sweep-angle!)
  (let ((new (flo:* (->flonum degrees) (flo:/ flo:pi 180.))))
    (if (not (flo:= new (arc-ink-%sweep-angle arc)))
	(begin
	  (set-arc-ink-%sweep-angle! arc new)
	  (drawing-damage arc)))))

(define (arc-ink-width ink)
  (guarantee-arc-ink ink 'arc-ink-width)
  (get-option ink 'LINE-WIDTH '()))

(define (set-arc-ink-width! ink width)
  (assert-glib-locked 'set-arc-ink-width!)
  (guarantee-arc-ink ink 'set-arc-ink-width!)
  (guarantee positive-fixnum? width 'set-arc-ink-width!)
  (without-interruption
   (lambda ()
     (if (set-option!? ink 'LINE-WIDTH (->flonum width))
	 (recache-arc-extent! ink)))))

(define (arc-ink-color ink)
  (guarantee-arc-ink ink 'arc-ink-color)
  (get-option ink 'OUTLINE '()))

(define (set-arc-ink-color! ink color)
  (assert-glib-locked 'set-arc-ink-color!)
  (guarantee-arc-ink ink 'set-arc-ink-color!)
  (let ((color (->color color 'set-arc-ink-color!)))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'OUTLINE color)
	   (drawing-damage ink))))))

(define (arc-ink-fill-color ink)
  (guarantee-arc-ink ink 'arc-ink-fill-color)
  (get-option ink 'FILL '()))

(define (set-arc-ink-fill-color! ink color)
  (assert-glib-locked 'set-arc-ink-fill-color!)
  (guarantee-arc-ink ink 'set-arc-ink-fill-color!)
  (let ((color (->color color 'set-arc-ink-fill-color!)))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'FILL color)
	   (drawing-damage ink))))))

(define-class (<text-ink> (constructor ()))
    (<draw-ink>))

(define-guarantee text-ink "a <text-ink>")

(define-generic text-ink-pango-layout (ink))

(define-method fix-ink-draw-callback ((ink <text-ink>) widget window cr area)
  (declare (ignore window area))
  (%trace2 ";drawing "ink" on "widget"\n")
  (let ((layout (text-ink-pango-layout ink)))
    (if layout
	(let ((view (fix-layout-view widget))
	      (rect (fix-ink-extent ink)))
	  (let ((x (fix:- (fix-rect-x rect) (fix-rect-x view)))
		(y (fix:- (fix-rect-y rect) (fix-rect-y view))))
	    (set-text-options! cr ink)
	    (cairo-move-to cr x y)
	    (cairo-show-pango-layout cr layout))))))

(define (set-text-options! cr ink)
  (for-each
    (lambda (entry)
      (let ((name (car entry))
	    (value (cdr entry)))
	(case name
	  ((COLOR) (cairo-set-source-color cr value)))))
    (draw-ink-options ink)))

(define (set-text-ink-position! ink x y)
  (assert-glib-locked 'set-text-ink-position!)
  (guarantee fixnum? x 'set-text-ink-position!)
  (guarantee fixnum? y 'set-text-ink-position!)
  (without-interruption
   (lambda ()
     (let ((rect (fix-ink-extent ink)))
       (if (not (and (fix:= x (fix-rect-x rect))
		     (fix:= y (fix-rect-y rect))))
	   (begin
	     (drawing-damage ink)
	     (set-fix-rect-position! rect x y)
	     (drawing-damage ink)))))))

(define-method fix-ink-move! ((ink <text-ink>) dx dy)
  (generic-fix-ink-move! ink dx dy))

(define (recache-text-extent! ink)
  (let ((layout (text-ink-pango-layout ink)))
    (pango-layout-get-pixel-extents
     layout
     (lambda (width height)
       (drawing-damage ink)
       (set-fix-rect-size! (fix-ink-extent ink) width height)
       (drawing-damage ink)))))

(define (text-ink-color ink)
  (guarantee-text-ink ink 'text-ink-color)
  (get-option ink 'COLOR '()))

(define (set-text-ink-color! ink color)
  (assert-glib-locked 'set-text-ink-color!)
  (guarantee-text-ink ink 'set-text-ink-color!)
  (let ((color (->color color 'set-text-ink-color!)))
    (without-interruption
     (lambda ()
       (if (set-option!? ink 'COLOR color)
	   (drawing-damage ink))))))

(define (text-ink-xy-to-index ink x y)
  (and (text-ink-pango-layout ink)
       (pango-layout-xy-to-index (text-ink-pango-layout ink)
				 (fix:- x (fix-rect-x (fix-ink-extent ink)))
				 (fix:- y (fix-rect-y (fix-ink-extent ink))))))

(define (with-text-ink-grapheme-rect ink index receiver)
  (and (text-ink-pango-layout ink)
       (pango-layout-index-to-pos (text-ink-pango-layout ink)
				  index receiver)))

(define (->pango-font-description spec operator)
  (cond ((and (alien? spec) (eq? '|PangoFontDescription| (alien/ctype spec)))
	 spec)
	((string? spec)
	 (or (hash-table-ref/default cached-font-descriptions spec #f)
	     (let ((alien (pango-font-description-from-string spec)))
	       (if (not alien)
		   (error:wrong-type-argument spec "PangoFontDescription"
					      operator))
	       (hash-table-set! cached-font-descriptions spec alien)
	       alien)))
	(else (error:wrong-type-argument spec "PangoFontDescription"
					 operator))))

(define cached-font-descriptions (make-string-hash-table))

(define (reset-font-descriptions!)
  (set! cached-font-descriptions (make-string-hash-table)))

(add-event-receiver! event:after-restore reset-font-descriptions!)

(define-class (<simple-text-ink> (constructor ()))
    (<text-ink>)

  ;; A Scheme string.  The content of the paragraph.
  (text define standard initial-value #f modifier set-simple-text-ink-%text!)

  ;; Corresponding PangoLayout.
  (pango-layout define standard initial-value #f))

(define-guarantee simple-text-ink "a <simple-text-ink>")

(define-method text-ink-pango-layout ((ink <simple-text-ink>))
  ;; Simply return the PangoLayout, which is only #f if no text has
  ;; (ever) been set.
  (simple-text-ink-pango-layout ink))

(define (set-simple-text-ink-text! ink widget text)
  ;; The TEXT string is shared.
  (assert-glib-locked 'set-simple-text-ink-text!)
  (guarantee-simple-text-ink ink 'set-simple-text-ink-text!)
  (guarantee-gtk-widget widget 'set-simple-text-ink-text!)
  (guarantee string? text 'set-simple-text-ink-text!)
  (without-interruption
   (lambda ()
     (let ((old (simple-text-ink-text ink)))
       (if (not (and old (string=? text old)))
	   (let ((layout (simple-text-ink-pango-layout ink)))
	     (if layout
		 (pango-layout-set-text layout text)
		 (let* ((desc (get-option ink 'FONT #f))
			(layout
			 (if desc
			     (let ((layout
				    (gtk-widget-create-pango-layout widget "")))
			       (pango-layout-set-font-description layout desc)
			       (pango-layout-set-text layout text)
			       layout)
			     (gtk-widget-create-pango-layout widget text))))
		   (set-simple-text-ink-pango-layout! ink layout)))
	     (set-simple-text-ink-%text! ink text)
	     (recache-text-extent! ink)))))))

(define (simple-text-ink-font ink)
  (guarantee-simple-text-ink ink 'simple-text-ink-font)
  (get-option ink 'FONT #f))

(define (set-simple-text-ink-font! ink font)
  (assert-glib-locked 'set-simple-text-ink-font!)
  (guarantee-simple-text-ink ink 'set-simple-text-ink-font!)
  (let ((new (->pango-font-description font 'set-simple-text-ink-font!)))
    (without-interruption
     (lambda ()
       (let ((layout (simple-text-ink-pango-layout ink)))
	 (if (and (set-option!? ink 'FONT new) layout)
	     (begin
	       (pango-layout-set-font-description layout new)
	       (recache-text-extent! ink))))))))

(define-class (<image-ink> (constructor ()))
    (<fix-ink>)
  ;; This slot is set to a <pixbuf> soon after loading has begun.
  (pixbuf define standard initial-value #f)
  ;; This slot is set to #f when the pixbuf has been successfully loaded.
  (loader define standard initializer make-pixbuf-loader))

(define-method initialize-instance ((ink <image-ink>))
  (call-next-method ink)
  (%trace ";(initialize-instance <image-ink>) "ink"\n")
  (let ((loader (image-ink-loader ink)))
    (set-pixbuf-loader-size-hook! loader (image-ink-size-prepared ink))
    (set-pixbuf-loader-pixbuf-hook! loader (image-ink-pixbuf-prepared ink))
    (set-pixbuf-loader-update-hook! loader (image-ink-pixbuf-updated ink))
    (set-pixbuf-loader-close-hook! loader (image-ink-pixbuf-loaded ink))))

(define (image-ink-size-prepared ink)
  (named-lambda (image-ink-size-prepared-handler width height)
    (%trace ";image-ink-size-prepared-handler "ink" "width" "height"\n")
    (let ((extent (fix-ink-extent ink)))
      (if (not (and (fix:= width (fix-rect-width extent))
		    (fix:= height (fix-rect-height extent))))
	  (begin
	    (drawing-damage ink)
	    (set-fix-rect-size! extent width height)
	    (drawing-damage ink))))))

(define (image-ink-pixbuf-prepared ink)
  (named-lambda (image-ink-pixbuf-prepared-handler pixbuf)
    (%trace ";image-ink-pixbuf-prepared-handler "ink" "pixbuf"\n")
    (set-image-ink-pixbuf! ink pixbuf)))

(define (image-ink-pixbuf-updated ink)
  (named-lambda (image-ink-pixbuf-updated-handler x y width height)
    (let* ((extent (fix-ink-extent ink))
	   (rect (make-fix-rect
		  (fix:+ x (fix-rect-x extent))
		  (fix:+ y (fix-rect-y extent))
		  width height)))
      (%trace ";image-ink-pixbuf-updated-handler "ink" "rect"\n")
      (drawing-damage ink rect))))

(define (image-ink-pixbuf-loaded ink)
  (named-lambda (image-ink-pixbuf-loaded-handler loader)
    (%trace ";image-ink-pixbuf-loaded-handler "ink" ("(image-ink-pixbuf ink)")"
	    " "(pixbuf-loader-error-message loader)"\n")
    (if (not (pixbuf-loader-error-message loader))
	(begin
	  (set-image-ink-loader! ink #f)
	  (gobject-unref! loader))
	(begin
	  ;; Hack the pixbuf with a "broken image" overlay?
	  ;;
	  ;; Leave the loader, with dead thread and closed
	  ;; input-port, for debugging purposes.
	  unspecific))))

(define-method fix-ink-draw-callback ((ink <image-ink>) widget window cr area)
  (declare (ignore window area))
  (%trace2 ";drawing "ink" on "widget"\n")
  (assert-glib-locked '(fix-ink-draw-callback <image-ink>))
  (let ((pixbuf (let ((p (image-ink-pixbuf ink)))
		  (if p (gobject-alien p) #f))))
    (if (and pixbuf (not (alien-null? pixbuf)))
	(let ((view (fix-layout-view widget))
	      (extent (fix-ink-extent ink)))
	  (let ((x. (->flonum (fix:- (fix-rect-x extent) (fix-rect-x view))))
		(y. (->flonum (fix:- (fix-rect-y extent) (fix-rect-y view)))))
	    (C-call "gdk_cairo_set_source_pixbuf" cr pixbuf x. y.)
	    (cairo-paint cr))))))

(define-method fix-ink-move! ((ink <image-ink>) dx dy)
  (generic-fix-ink-move! ink dx dy))

(define (make-image-ink-from-file filename)
  (let ((ink (make-image-ink)))
    (load-pixbuf-from-file (image-ink-loader ink) filename)
    ink))

(define (set-image-ink! ink x y)
  (guarantee fixnum? x 'set-image-ink-position!)
  (guarantee fixnum? y 'set-image-ink-position!)
  (set-fix-ink-%position! ink x y))

(define-class (<surface-ink> (constructor () (width height)))
    (<fix-ink>)

  ;; Cairo Image Surface -- a |cairo_surface_t| alien.
  (surface define standard)

  ;; For use by the glib thread only.
  (exposed define standard initial-value #f))

(define-method initialize-instance ((ink <surface-ink>) width height)
  (call-next-method ink)
  (set-fix-rect-size! (fix-ink-extent ink) width height)
  (set-surface-ink-surface! ink (cairo-image-surface-create width height)))

(define-method fix-ink-draw-callback ((ink <surface-ink>) widget window cr area)
  (declare (ignore area))
  (%trace ";drawing "ink" on "widget"\n")
  (let ((view (fix-layout-view widget))
	(extent (fix-ink-extent ink))
	(surface (get-exposed ink window)))
    (let ((x (fix:- (fix-rect-x extent) (fix-rect-x view)))
	  (y (fix:- (fix-rect-y extent) (fix-rect-y view))))
      (cairo-set-source-surface cr surface x y)
      (cairo-paint cr))))

(define (get-exposed ink window)
  (assert-glib-locked 'get-exposed)
  (or (surface-ink-exposed ink)
      (let ((extent (fix-ink-extent ink))
	    (scale (C-call "gdk_window_get_scale_factor" window)))
	(let ((width (fix:* scale (fix-rect-width extent)))
	      (height (fix:* scale (fix-rect-height extent))))
	  (let* ((surface (gdk-window-create-similar-surface window
							     width height))
		 (cr (cairo-create surface)))
	    (cairo-set-source-surface cr (surface-ink-surface ink) 0 0)
	    (cairo-paint cr)
	    (cairo-destroy cr)
	    (set-surface-ink-exposed! ink surface)
	    surface)))))

(define (surface-ink-flush ink)
  (let ((surface (surface-ink-surface ink)))
    (cairo-surface-flush surface)
    (with-glib-lock
     (lambda ()
       (let ((exposed (surface-ink-exposed ink)))
	 (if exposed
	     (let ((cr (cairo-create exposed)))
	       (cairo-set-source-surface cr surface 0 0)
	       (cairo-paint cr)
	       (cairo-destroy cr)
	       (cairo-surface-flush exposed)))
	 (drawing-damage ink))))))

(define (set-surface-ink-position! ink x y)
  (set-fix-rect-position! (fix-ink-extent ink) x y))

;;;; Fixnum Rectangles

(define-structure (fix-rect (constructor make-fix-rect (#!optional x y width height))
			(copier)
			(print-procedure 
			 (bracketed-print-method 'fix-rect
			  (lambda (rect port)
			    (write-string " " port)
			    (write-string (fix-rect-string rect) port)))))
  (x #f) (y #f) (width #f) (height #f))

(define (fix-rect-string rect)
  (define-integrable (S a) (write-to-string a))
  (string-append (S (fix-rect-x rect))","(S (fix-rect-y rect))
		 " "(S (fix-rect-width rect))"x"(S (fix-rect-height rect))))

(define-integrable-operator (make-fix-rect-from-bounds min-x min-y max-x max-y)
  (make-fix-rect min-x min-y (fix:- max-x min-x) (fix:- max-y min-y)))

(define-integrable-operator (set-fix-rect! rect x y width height)
  (set-fix-rect-x! rect x)
  (set-fix-rect-y! rect y)
  (set-fix-rect-width! rect width)
  (set-fix-rect-height! rect height))

(define-integrable-operator (set-fix-rect-bounds! rect min-x max-x min-y max-y)
  (set-fix-rect-x! rect min-x)
  (set-fix-rect-y! rect min-y)
  (set-fix-rect-width! rect (fix:- max-x min-x))
  (set-fix-rect-height! rect (fix:- max-y min-y)))

(define-integrable-operator (set-fix-rect-position! rect x y)
  (set-fix-rect-x! rect x)
  (set-fix-rect-y! rect y))

(define-integrable-operator (fix-rect-move! rect dx dy)
  (set-fix-rect-x! rect (fix:+ (fix-rect-x rect) dx))
  (set-fix-rect-y! rect (fix:+ (fix-rect-y rect) dy)))

(define-integrable-operator (set-fix-rect-size! rect width height)
  (set-fix-rect-width! rect width)
  (set-fix-rect-height! rect height))

(define-integrable-operator (fix-rect-nominal? rect)
  (and (fixnum? (fix-rect-x rect))
       (fixnum? (fix-rect-y rect))
       (fixnum? (fix-rect-width rect))
       (fixnum? (fix-rect-height rect))))

;;; The rest of these procedures assume a "nominal" rectangle.

(define-integrable-operator (fix-rect-max-y rect)
  (fix:+ (fix-rect-y rect) (fix-rect-height rect)))

(define-integrable-operator (fix-rect-max-x rect)
  (fix:+ (fix-rect-x rect) (fix-rect-width rect)))

(define-integrable fix-rect-min-x fix-rect-x)
(define-integrable fix-rect-min-y fix-rect-y)

(define-integrable-operator (with-fix-rect-bounds rect receiver)
  ;; Tail-calls RECEIVER with the RECT's minx, maxx, miny and maxy (in
  ;; that order).  Assumes RECT is nominal.
  (let ((x (fix-rect-x rect))
	(y (fix-rect-y rect))
	(width (fix-rect-width rect))
	(height (fix-rect-height rect)))
    (receiver x (fix:+ x width) y (fix:+ y height))))

(define-integrable-operator (with-fix-rect rect receiver)
  (receiver (fix-rect-x rect) (fix-rect-y rect) (fix-rect-width rect) (fix-rect-height rect)))

(define-integrable-operator (copy-fix-rect! target source)
  (set-fix-rect-x! target (fix-rect-x source))
  (set-fix-rect-y! target (fix-rect-y source))
  (set-fix-rect-width! target (fix-rect-width source))
  (set-fix-rect-height! target (fix-rect-height source)))

(define-integrable-operator (point-in-fix-rect? x y rect)
  (with-fix-rect-bounds rect
   (lambda (min-x max-x min-y max-y)
     (and (fix:<= min-x x) (fix:<= x max-x)
	  (fix:<= min-y y) (fix:<= y max-y)))))

(define-integrable-operator (fix-rect-at-point? rect x y)
  (and (fix:= x (fix-rect-x rect))
       (fix:= y (fix-rect-y rect))))

(define-integrable-operator (fix-rect-intersect? rect1 rect2)
  ;; Useful when you do not need to cons a new rect.
  (with-fix-rect-bounds rect1
   (lambda (min-x1 max-x1 min-y1 max-y1)
     (with-fix-rect-bounds rect2
      (lambda (min-x2 max-x2 min-y2 max-y2)
	(cond ((fix:< max-x1 min-x2) #f)
	      ((fix:< max-y1 min-y2) #f)
	      ((fix:< max-x2 min-x1) #f)
	      ((fix:< max-y2 min-y1) #f)
	      (else #t)))))))

(define (fix-rect-intersection rect1 rect2)
  ;; Returns #f if RECT1 and RECT2 do not intersect, else returns a
  ;; new rect -- the intersection.  Assumes both rectangles are
  ;; nominal.
  (with-fix-rect-bounds rect1
   (lambda (min-x1 max-x1 min-y1 max-y1)
     (with-fix-rect-bounds rect2
      (lambda (min-x2 max-x2 min-y2 max-y2)
	(cond ((fix:< max-x1 min-x2) #f)
	      ((fix:< max-y1 min-y2) #f)
	      ((fix:< max-x2 min-x1) #f)
	      ((fix:< max-y2 min-y1) #f)
	      (else
	       (let ((min-x (fix:max min-x1 min-x2))
		     (min-y (fix:max min-y1 min-y2))
		     (max-x (fix:min max-x1 max-x2))
		     (max-y (fix:min max-y1 max-y2)))
		 (make-fix-rect min-x min-y
			    (fix:- max-x min-x)
			    (fix:- max-y min-y))))))))))

(define (window-intersection window item)
  ;; Returns #f if WINDOW and ITEM do not intersect, else returns a
  ;; new rect -- the intersection *translated* to WINDOW's coords.
  ;; Assumes both rectangles are nominal.
  (with-fix-rect-bounds window
   (lambda (window-x-start window-x-end window-y-start window-y-end)
     (with-fix-rect-bounds item
      (lambda (item-x-start item-x-end item-y-start item-y-end)
	(cond ((fix:< window-x-end item-x-start) #f)
	      ((fix:< window-y-end item-y-start) #f)
	      ((fix:< item-x-end window-x-start) #f)
	      ((fix:< item-y-end window-y-start) #f)
	      (else
	       (let ((x (fix:max window-x-start item-x-start))
		     (y (fix:max window-y-start item-y-start))
		     (x-end (fix:min window-x-end item-x-end))
		     (y-end (fix:min window-y-end item-y-end)))
		 (make-fix-rect (fix:- x window-x-start)
			    (fix:- y window-y-start)
			    (fix:- x-end x)
			    (fix:- y-end y))))))))))

(define (fix-rect-union! rect1 rect2)
  (with-fix-rect-bounds rect1
   (lambda (min-x1 max-x1 min-y1 max-y1)
     (with-fix-rect-bounds rect2
      (lambda (min-x2 max-x2 min-y2 max-y2)
	(let ((x (fix:min min-x1 min-x2))
	      (y (fix:min min-y1 min-y2)))
	  (set-fix-rect! rect1
			 x y
			 (fix:- (fix:max max-x1 max-x2) x)
			 (fix:- (fix:max max-y1 max-y2) y))))))))

(define (fix-rect-contains? rect1 rect2)
  ;; True if RECT2 is wholly contained within RECT1.
  (with-fix-rect-bounds rect1
   (lambda (min-x1 max-x1 min-y1 max-y1)
     (with-fix-rect-bounds rect2
      (lambda (min-x2 max-x2 min-y2 max-y2)
	(and (fix:<= min-x1 min-x2) (fix:<= max-x2 max-x1)
	     (fix:<= min-y1 min-y2) (fix:<= max-y2 max-y1)))))))

(define (gdk-rectangle #!optional x y width height)
  (if (not (default-object? x)) (guarantee fixnum? x 'gdk-rectangle))
  (if (not (default-object? y)) (guarantee fixnum? y 'gdk-rectangle))
  (if (not (default-object? width)) (guarantee-size width 'gdk-rectangle))
  (if (not (default-object? height)) (guarantee-size height 'gdk-rectangle))
  (let ((alien (malloc (C-sizeof "GdkRectangle") '|GdkRectangle|)))
    (if (default-object? x) alien
	(begin
	  (C->= alien "GdkRectangle x" x)
	  (if (default-object? y) alien
	      (begin
		(C->= alien "GdkRectangle y" y)
		(if (default-object? width) alien
		    (begin
		      (C->= alien "GdkRectangle width" width)
		      (if (default-object? height) alien
			  (begin
			    (C->= alien "GdkRectangle height" height)
			    alien))))))))))

(define (gdk-rectangle-from-rect rect)
  (gdk-rectangle (fix-rect-x rect) (fix-rect-y rect)
		 (fix-rect-width rect) (fix-rect-height rect)))


(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace? (outf-error ARGS ...)))))

(define %trace2? #f)

(define-syntax %trace2
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace2? (outf-error ARGS ...)))))