#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; GDK Objects
;;; package: (gdk)

(define-integrable-operator (guarantee-gdk-window object operator)
  (if (not (and (alien? object) (eq? '|GdkWindow| (alien/ctype object))))
      (error:wrong-type-argument object "a GdkWindow address" operator)))

(define (gdk-window-create-similar-surface window width height)
  (assert-glib-locked 'gdk-window-create-similar-surface)
  (let ((surface (make-alien '|cairo_surface_t|))
	(copy (make-alien '|cairo_surface_t|)))
    (add-glib-cleanup surface (make-cairo-surface-cleanup copy))
    (C-call "gdk_window_create_similar_surface" copy
	    window (C-enum "CAIRO_CONTENT_COLOR") width height)
    (copy-alien-address! surface copy)
    (check-cairo-surface-status surface)
    surface))

(define (gdk-window-create-similar-image-surface window width height scale)
  (assert-glib-locked 'gdk-window-create-similar-image-surface)
  (let ((surface (make-alien '|cairo_surface_t|))
	(copy (make-alien '|cairo_surface_t|)))
    (add-glib-cleanup surface (make-cairo-surface-cleanup copy))
    (C-call "gdk_window_create_similar_image_surface" copy
	    window (C-enum "CAIRO_FORMAT_RGB24") width height scale)
    (copy-alien-address! surface copy)
    (check-cairo-surface-status surface)
    surface))

;;; GdkPixbufLoaders

(define-class (<pixbuf-loader> (constructor ()))
     (<gobject>)
  (port define standard initial-value #f)
  (thread define standard initial-value #f)
  (size define standard initial-value #f)
  (pixbuf define standard initial-value #f)
  (error-message define standard initial-value #f)
  (closed? define standard initial-value #f)
  (size-hook define standard initial-value #f
	     modifier %set-pixbuf-loader-size-hook!)
  (pixbuf-hook define standard initial-value #f
	       modifier %set-pixbuf-loader-pixbuf-hook!)
  (update-hook define standard initial-value #f)
  (close-hook define standard initial-value #f
	      modifier %set-pixbuf-loader-close-hook!))

(define-class (<pixbuf> (constructor ()))
    (<gobject>))

(define-method initialize-instance ((pixbuf <pixbuf>))
  (call-next-method pixbuf)
  (set-alien/ctype! (gobject-alien pixbuf) '|GdkPixbuf|))

(define-method initialize-instance ((loader <pixbuf-loader>))
  (call-next-method loader)
  (assert-glib-locked '(initialize-instance <pixbuf-loader>))
  (C-call "gdk_pixbuf_loader_new" (gobject-alien loader))
  (g-signal-connect loader (C-callback "size_prepared")
		    pixbuf-loader-size-prepared)
  (g-signal-connect loader (C-callback "area_prepared")
		    pixbuf-loader-area-prepared)
  (g-signal-connect loader (C-callback "area_updated")
		    pixbuf-loader-area-updated))

(define (pixbuf-loader-size-prepared loader width height)
  (%trace "; pixbuf-loader-size-prepared "loader" "width" "height"\n")
  (assert-glib-locked 'pixbuf-loader-size-prepared)
  (let ((size (pixbuf-loader-size loader)))
    (if size (outf-error ";pixbuf loader already has a size: "loader"\n"))
    (set-pixbuf-loader-size! loader (cons width height))
    (let ((receiver (pixbuf-loader-size-hook loader)))
      (if receiver (receiver width height)))))

(define (pixbuf-loader-area-prepared loader)
  (%trace "; pixbuf-loader-area-prepared "loader"\n")
  (assert-glib-locked 'pixbuf-loader-area-prepared)
    (let* ((alien (gobject-alien loader))
	   (pixbuf (let ((p (pixbuf-loader-pixbuf loader)))
		     (if p
			 (error "Pixbuf loader already has a pixbuf:" loader)
			 (make-pixbuf))))
	   (pixbuf-alien (gobject-alien pixbuf)))
      (C-call "gdk_pixbuf_loader_get_pixbuf" pixbuf-alien alien)
      (C-call "g_object_ref" #f pixbuf-alien)
      (set-pixbuf-loader-pixbuf! loader pixbuf)
      (let ((receiver (pixbuf-loader-pixbuf-hook loader)))
	(if receiver (receiver pixbuf)))))

(define (pixbuf-loader-area-updated loader x y width height)
  (%trace "; pixbuf-loader-area-updated "loader" "x","y" "width"x"height"\n")
  (assert-glib-locked 'pixbuf-loader-area-updated)
  (let ((receiver (pixbuf-loader-update-hook loader)))
    (if receiver (receiver x y width height))))

(define (load-pixbuf-from-port loader input-port)
  (if (pixbuf-loader-port loader)
      (error "Pixbuf loader has already started:" loader))
  (set-pixbuf-loader-port! loader input-port)
  (detach-thread (create-pixbuf-loader-thread loader))
  unspecific)

(define (create-pixbuf-loader-thread loader)
  (create-thread
      #f
    (named-lambda (load-pixbuf)
      (%trace "; "loader" started in "(current-thread)"\n")
      (set-pixbuf-loader-thread! loader (current-thread))
      (let ((port (pixbuf-loader-port loader))
	    (alien (gobject-alien loader))
	    (gerror* (make-gerror-pointer))
	    (buffer (make-bytevector 4200)))
	(C->= gerror* "* GError" 0)

	(define (note-done)
	  (assert-without-interruption 'load-pixbuf)
	  (gerror-pointer-free gerror*)
	  (set-pixbuf-loader-closed?! loader #t)
	  (close-input-port port)
	  (%trace "; "loader" closed by "(current-thread)"\n")
	  (let ((proc (pixbuf-loader-close-hook loader)))
	    (if proc
		(proc loader)))
	  unspecific)

	(define (note-error)
	  (let* ((gerror (C-> gerror* "* GError"))
		 (message (or (and (not (alien-null? gerror))
				   (c-peek-cstring
				    (C-> gerror "GError message")))
			      "GError pointer not set.")))
	    (set-pixbuf-loader-error-message! loader message))
	  (note-done))

	(define-integrable (with-glib-atom thunk)
	  (with-glib-lock
	   (lambda ()
	     (without-interruption thunk))))

	(let loop ()
	  (let ((n (read-bytevector! buffer port)))
	    (cond ((eof-object? n)
		   (with-glib-atom
		    (lambda ()
		      (if (fix:zero? (C-call "gdk_pixbuf_loader_close"
					     alien gerror*))
			  (note-error)
			  (note-done)))))
		  ((or (not (fix:fixnum? n))
		       (fix:zero? n))
		   (with-glib-atom note-error))
		  ((not (fix:zero?
			 (with-glib-lock
			  (lambda ()
			    (C-call "gdk_pixbuf_loader_write"
				    alien buffer n gerror*)))))
		   (loop))
		  (else
		   (with-glib-atom note-error)))))))))

(define (make-gerror-pointer)
  (with-glib-lock
   (lambda ()
     (let ((alien (make-alien '(* |GError|)))
	   (copy  (make-alien '(* |GError|))))
       (add-glib-cleanup alien (make-gerror-pointer-cleanup copy))
       (C-call "g_try_malloc0" copy (C-sizeof "* GError"))
       (copy-alien-address! alien copy)
       (error-if-null alien "Could not create:" alien)
       alien))))

(define (make-gerror-pointer-cleanup copy)
  (named-lambda (cleanup-gerror-pointer)
    (assert-glib-locked 'cleanup-gerror-pointer)
    (assert-without-interruption 'cleanup-gerror-pointer)
    (if (not (alien-null? copy))
	(let ((gerror (make-alien '|GError|)))
	  (C-> copy "* GError" gerror)
	  (if (not (alien-null? gerror))
	      (C-call "g_error_free" gerror))
	  (C-call "g_free" copy)
	  (alien-null! copy)))))

(define (gerror-pointer-free gerror*)
  (assert-glib-locked 'gerror-pointer-free)
  (assert-without-interruption 'gerror-pointer-free)
  (if (not (alien-null? gerror*))
      (begin
	(execute-glib-cleanup gerror*)
	(alien-null! gerror*))))

(define (load-pixbuf-from-file loader filename)
  (load-pixbuf-from-port
   loader (open-binary-input-file (->namestring (->truename filename)))))

(define (set-pixbuf-loader-size-hook! loader receiver)
  (assert-glib-locked 'set-pixbuf-loader-size-hook!) ; serialize with loader
  (%set-pixbuf-loader-size-hook! loader receiver)
  (let ((size (pixbuf-loader-size loader)))
    (if size (receiver (car size) (cdr size)))))

(define (set-pixbuf-loader-pixbuf-hook! loader receiver)
  (assert-glib-locked 'set-pixbuf-loader-pixbuf-hook!) ; serialize with loader
  (%set-pixbuf-loader-pixbuf-hook! loader receiver)
  (let ((pixbuf (pixbuf-loader-pixbuf loader)))
    (if pixbuf (receiver pixbuf))))

(define (set-pixbuf-loader-close-hook! loader thunk)
  (assert-glib-locked 'set-pixbuf-loader-close-hook!) ; serialize with loader
  (%set-pixbuf-loader-close-hook! loader thunk)
  (if (pixbuf-loader-closed? loader)
      (thunk)))

;;; GdkDisplays

(define-record-type <gdk-display>
    (make-gdk-display alien atoms clipboard queue callback-idv)
    gdk-display?
  (alien gdk-display/alien)
  (atoms gdk-display/atoms set-gdk-display/atoms!)
  (clipboard gdk-display/clipboard set-gdk-display/clipboard!)
  (queue gdk-display/queue)
  (callback-idv gdk-display/callback-idv))

(define displays (make-key-weak-eqv-hash-table))

(define (get-gdk-display alien)
  (let ((bignum (alien/address alien)))
    (or (hash-table-ref/default displays bignum #f)
	(let ((alien (copy-alien alien))
	      (callback-idv (vector #f))
	      (queue (make-thread-queue 1)))
	  (let ((display (make-gdk-display alien '() #f queue callback-idv)))
	    (add-glib-cleanup display (make-gdk-display-cleanup callback-idv))
	    (hash-table-set! displays bignum display)
	    display)))))

(define (make-gdk-display-cleanup callback-idv)
  (named-lambda (gdk-display-cleanup)
    (cleanup-callback callback-idv)))

(define (cleanup-callback callback-idv)
  (let ((id (vector-ref callback-idv 0)))
    (if id
	(begin
	  (de-register-c-callback id)
	  (vector-set! callback-idv 0 #f)))))

(define (clipboard display)
  (assert-glib-locked 'clipboard)
  (or (gdk-display/clipboard display)
      (let ((atom (get-atom display '|CLIPBOARD|))
	    (gdkdisplay (gdk-display/alien display))
	    (clipboard (make-alien '|GtkClipboard|)))
	(set-gdk-display/clipboard! display clipboard)
	(C-call "gtk_clipboard_get_for_display" clipboard gdkdisplay atom)
	clipboard)))

(define (get-atom display symbol)
  (assert-glib-locked 'get-atom)
  (let ((entry (assq symbol (gdk-display/atoms display))))
    (if entry
	(cdr entry)
	(let ((atom (make-alien '(struct |_GdkAtom|))))
	  (C-call "gdk_atom_intern" atom
		  (string->utf8 (symbol->string symbol)) 0)
	  (set-gdk-display/atoms! display
				  (cons (cons symbol atom)
					(gdk-display/atoms display)))
	  atom))))

(define (gdk-display-set-clipboard-text display string)
  (%trace "; gdk-display-set-clipboard-text "display"\n")
  (assert-glib-locked 'gdk-display-set-clipboard-text)
  (let ((string-bv (string->utf8 string)))
    (C-call "gtk_clipboard_set_text"
	     (clipboard display)
	     string-bv (bytevector-length string-bv))))

(define (gdk-display-get-clipboard-text display msec)
  (%trace "; gdk-display-get-clipboard-text "display" "msec"\n")
  (assert-glib-locked 'gdk-display-get-clipboard-text)
  (if (vector-ref (gdk-display/callback-idv display) 0)
      (error "Operation pending:" display))
  (let ((queue (gdk-display/queue display))
	(callback-idv (gdk-display/callback-idv display)))
    (%trace "; gdk-display-get-clipboard-text registering\n")
    (let ((callback-id (make-text-callback-id queue)))
      (vector-set! callback-idv 0 callback-id)
      (thread-queue/empty! queue)
      (C-call "gtk_clipboard_request_text"
	      (clipboard display)
	      (C-callback "receive_clipboard_text")
	      callback-id)
      (%trace "; gdk-display-get-clipboard-text waiting\n")
      (let ((text (thread-queue/dequeue-no-hang! queue msec)))
	(%trace "; gdk-display-get-clipboard-text finishing\n")
	(cleanup-callback callback-idv)
	(if (string? text)
	    text
	    (error "Operation failed:" display))))))

(define (make-text-callback-id queue)
  (C-callback
   (named-lambda (gdk-display-get-clipboard-text-callback clipboard char*)
     (declare (ignore clipboard))
     (if (alien-null? char*)
	 (queue! queue #t)
	 (queue! queue (c-peek-cstring char*))))))

(define (queue! queue value)
  (thread-queue/queue! queue value)
  (maybe-yield-glib))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace? (outf-error ARGS ...)))))