/* -*-C-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

*/

/* The ScmWidget, represented in Scheme by a <scm-widget>. */

/* #include <mit-scheme.h> */
#include "gtk-shim.h"
/* #include <gtkadjustment.h> */
/* #include <gtk/gtkscrollable.h> */

enum {
  PROP_0,
  PROP_HADJUSTMENT,
  PROP_VADJUSTMENT,
  PROP_HSCROLL_POLICY,
  PROP_VSCROLL_POLICY
};

static void scm_widget_set_property (GObject *object, guint prop_id,
				     const GValue *value, GParamSpec *pspec);
static void scm_widget_get_property (GObject *object, guint prop_id,
				     GValue *value, GParamSpec *pspec);
static void scm_widget_realize (GtkWidget *widget);
static void scm_widget_get_preferred_width (GtkWidget *widget,
					    gint *minimum_size,
					    gint *natural_size);
static void scm_widget_get_preferred_height (GtkWidget *widget,
					     gint *minimum_size,
					     gint *natural_size);

static guint set_scroll_adjustments_signal_id;

static void
marshal_VOID__OBJECT_OBJECT (GClosure     *closure,
			     GValue       *return_value G_GNUC_UNUSED,
			     guint         n_param_values,
			     const GValue *param_values,
			     gpointer      invocation_hint G_GNUC_UNUSED,
			     gpointer      marshal_data)
{
  typedef void (*MarshalFunc) (gpointer     data1,
				gpointer     arg_1,
				gpointer     arg_2,
				gpointer     data2);
  register MarshalFunc callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_get_object (param_values + 0);
    }
  else
    {
      data1 = g_value_get_object (param_values + 0);
      data2 = closure->data;
    }
  callback = (MarshalFunc) (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_value_get_object (param_values + 1),
            g_value_get_object (param_values + 2),
            data2);
}

G_DEFINE_TYPE_WITH_CODE (ScmWidget, scm_widget, GTK_TYPE_WIDGET,
                         G_IMPLEMENT_INTERFACE (GTK_TYPE_SCROLLABLE, NULL))

static void
scm_widget_init (ScmWidget *widget)
{
  widget->minimum_width = 0;
  widget->minimum_height = 0;
  widget->natural_width = 0;
  widget->natural_height = 0;
  widget->hadjustment = NULL;
  widget->vadjustment = NULL;
  widget->hscroll_policy = GTK_SCROLL_NATURAL;
  widget->vscroll_policy = GTK_SCROLL_NATURAL;
}

static void
scm_widget_class_init (ScmWidgetClass *klass)
{
  GObjectClass *gobject_class;
  GtkWidgetClass *widget_class;

  gobject_class = G_OBJECT_CLASS (klass);
  widget_class = (GtkWidgetClass*) klass;

  gobject_class->set_property = scm_widget_set_property;
  gobject_class->get_property = scm_widget_get_property;

  widget_class->realize = scm_widget_realize;
  widget_class->get_preferred_width = scm_widget_get_preferred_width;
  widget_class->get_preferred_height = scm_widget_get_preferred_height;

  gtk_widget_class_set_accessible_role (widget_class, ATK_ROLE_VIEWPORT);

  /* GtkScrollable implementation */
  g_object_class_override_property
    (gobject_class, PROP_HADJUSTMENT,    "hadjustment");
  g_object_class_override_property
    (gobject_class, PROP_VADJUSTMENT,    "vadjustment");
  g_object_class_override_property
    (gobject_class, PROP_HSCROLL_POLICY, "hscroll-policy");
  g_object_class_override_property
    (gobject_class, PROP_VSCROLL_POLICY, "vscroll-policy");

  set_scroll_adjustments_signal_id =
    g_signal_new ("set_scroll_adjustments",
		  G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		  0, NULL, NULL, /* class_offset, accumulator, accu_data */
		  marshal_VOID__OBJECT_OBJECT,
		  G_TYPE_NONE, 2,
		  GTK_TYPE_ADJUSTMENT,
		  GTK_TYPE_ADJUSTMENT);
}

static void
scm_widget_set_property (GObject *object, guint prop_id,
			 const GValue *value, GParamSpec *pspec)
{
  ScmWidget *widget = SCM_WIDGET (object);

  switch (prop_id)
    {
    case PROP_HADJUSTMENT:
      {
	GtkAdjustment *adjustment = g_value_get_object (value);
	if (adjustment != widget->hadjustment)
	  {
	    widget->hadjustment = adjustment;
	    g_signal_emit (widget, set_scroll_adjustments_signal_id,
			   0, adjustment, widget->vadjustment);
	  }
      }
      break;
    case PROP_VADJUSTMENT:
      {
	GtkAdjustment *adjustment = g_value_get_object (value);
	if (adjustment != widget->vadjustment)
	  {
	    widget->vadjustment = adjustment;
	    g_signal_emit (widget, set_scroll_adjustments_signal_id,
			   0, widget->hadjustment, adjustment);
	  }
      }
      break;
    case PROP_HSCROLL_POLICY:
      {
	guint policy = g_value_get_enum (value);
	if (policy != widget->hscroll_policy)
	  {
	    widget->hscroll_policy = policy;
	    gtk_widget_queue_resize (GTK_WIDGET (widget));
	  }
      }
      break;
    case PROP_VSCROLL_POLICY:
      {
	guint policy = g_value_get_enum (value);
	if (policy != widget->vscroll_policy)
	  {
	    widget->vscroll_policy = policy;
	    gtk_widget_queue_resize (GTK_WIDGET (widget));
	  }
      }
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
scm_widget_get_property (GObject         *object,
			   guint            prop_id,
			   GValue          *value,
			   GParamSpec      *pspec)
{
  ScmWidget *widget = SCM_WIDGET (object);

  switch (prop_id)
    {
    case PROP_HADJUSTMENT:
      g_value_set_object (value, widget->hadjustment);
      break;
    case PROP_VADJUSTMENT:
      g_value_set_object (value, widget->vadjustment);
      break;
    case PROP_HSCROLL_POLICY:
      g_value_set_enum (value, widget->hscroll_policy);
      break;
    case PROP_VSCROLL_POLICY:
      g_value_set_enum (value, widget->vscroll_policy);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget *
scm_widget_new (void)
{
  return ((GtkWidget*) g_object_new (GTK_TYPE_SCMWIDGET, NULL));
}

static void
scm_widget_realize (GtkWidget* widget)
{
  /* The default GtkWidget method expects !gtk_widget_get_has_window,
     so it is replaced by this near no-op.  ScmWidget realization
     actually happens during REALIZE signal emission. */

  gtk_widget_set_realized (widget, TRUE);
}

static void
scm_widget_get_preferred_width (GtkWidget *widget,
				gint *minimum_size,
				gint *natural_size)
{
  ScmWidget *scm_widget = SCM_WIDGET (widget);

  if (minimum_size)
    *minimum_size = scm_widget->minimum_width;

  if (natural_size)
    *natural_size = scm_widget->natural_width;
}

static void
scm_widget_get_preferred_height (GtkWidget *widget,
				gint *minimum_size,
				gint *natural_size)
{
  ScmWidget *scm_widget = SCM_WIDGET (widget);

  if (minimum_size)
    *minimum_size = scm_widget->minimum_height;

  if (natural_size)
    *natural_size = scm_widget->natural_height;
}
