#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; SWAT adapter.
;;; package: (gtk swat)

;;;; SWAT Widgets

;;; These specializations of <gtk-widget> subclasses implement SWAT's
;;; widgets.

(define-class <swat-widget>
    (<gtk-widget>)

  ;; This flag is set by a realize signal handler (or fix-widget-realize).
  (realized? define standard initial-value #f)

  ;; An alist of SWAT's configuration settings (fonts, colors, etc.).
  (options define standard initial-value '())

  ;; A pair (list of pairs?): (reason . thunk).
  (on-death define standard initial-value #f))

(define-method initialize-instance ((widget <swat-widget>) . args)
  (%trace "(initialize-instance <swat-widget>) "widget" "args)
  (apply call-next-method widget args)
  ;; Connect to the "realize" signal to apply options like colors.
  ;; NOT replacing <fix-widget>'s realize callback; the
  ;; fix-widget-realize-callback method will apply the options.
  (if (not (fix-widget? widget))
      (set-gtk-widget-realize-callback! widget realize-options)))

(define-method gtk-widget-destroy-callback ((object <swat-widget>))
  (call-next-method object)
  (assert-glib-locked '(gtk-widget-destroy-callback <swat-widget>))
  (let ((on-death (swat-widget-on-death object)))
    (if on-death
	(begin
	  (set-swat-widget-on-death! object #f)
	  (%trace "on-death "object": "on-death)
	  ((cdr on-death))))))

(define-class (<swat-button> (constructor ()))
    (<swat-widget> <gtk-button>))

;; <swat-checkbutton>s can have a callback AND/or a -variable.  These
;; could be two signal handlers connected to the "toggled" signal, but
;; Gtk does not support more than one signal handler per signal name.
;; Instead, two slots keep track of the two possible callbacks, and
;; the "toggled" signal is always connected (even if neither a
;; callback nor a variable is configured).

(define-class (<swat-checkbutton> (constructor ()))
    (<swat-widget> <gtk-check-button>)

  ;; From (set-callback! checkbutton (lambda () ...))
  (swat-callback define standard initial-value #f)

  ;; From (make-checkbutton -variable ...)
  (swat-variable define standard initial-value #f))

;; <swat-label>s are actually GtkFrames with no "label", whose GtkBin
;; child is a GtkLabel... so a little confusing.

(define-class (<swat-label> (constructor ()))
    (<swat-widget> <gtk-frame>))

(define-method initialize-instance ((frame <swat-label>))
  (%trace "(initialize-instance <swat-label>) "frame)
  (call-next-method frame "")
  (with-glib-lock
   (lambda () (gtk-container-add frame (gtk-label-new "")))))

(define (set-label-relief! label relief)
  (let ((gtk-shadow-type (relief->gtk-shadow-type relief)))
    (with-glib-lock
     (lambda () (gtk-frame-set-shadow-type label gtk-shadow-type)))))

(define (relief->gtk-shadow-type relief)
  (case relief
    ((raised) 'out)
    ((sunken) 'in)
    ((flat) 'none)
    ((ridge) 'etched-out)
    ((solid) (warn "unimplemented relief" relief))
    ((groove) 'etched-in)
    (else (warn "unexpected relief" relief))))

;;;; SWAT canvases

;;; Implemented by a specialized <fix-layout> widget whose drawing
;;; consists of <swat-ink>s.  Swat inks co-operate as "items" in
;;; <swat-group>s.
;;;
;;; The implementation chosen here removes a group's items from the
;;; drawing's display-list, keeping them in the group's item list, and
;;; ensuring that the group's ink-extent encompasses all of its items'
;;; ink-extents.  The layout draw callback can then skip entire
;;; groups (assuming they are relatively compact) or descend and, in
;;; the group draw callback, perform a similar search-and-draw
;;; among the group's items, recursively.
;;;
;;; To make the group ink-extent guarantee, all of a group's items
;;; must be <swat-ink>s, which update their group's extent
;;; (recursively) whenever their extents change.
;;;
;;; Updating a group's extent generally means computing a new union of
;;; all item ink-extents.  This can be expensive for a large group
;;; when items change relative to the rest.  This is presumably rare.
;;; Once created, a group is likely only moved or hidden, with
;;; predictable effects on the group ink-extent (and no need to
;;; re-union-all).  For large groups of widely scattered objects, a
;;; different group implementation would be more appropriate, e.g. one
;;; that keeps its items in the drawing display-list and perhaps does
;;; not even appear there itself...

(define-class (<swat-canvas> (constructor () (width height)))
    (<swat-widget> <fix-layout>)

  ;; An alist of (event-type . modifiers) x SWAT event handler procedures.
  (swat-handlers define standard initial-value '()))

(define-method initialize-instance ((canvas <swat-canvas>) width height)
  (%trace "(initialize-instance <swat-canvas>) "canvas" "width"x"height)
  (call-next-method canvas width height '())
  (set-fix-layout-drawing! canvas (with-glib-lock make-fix-drawing) 0 0))

(define (set-swat-canvas-handler! canvas type.modifiers handler)
  ;; type.modifiers is e.g. (press 3 control), (double-press 1),
  ;; or (motion button1).
  (let* ((handlers (swat-canvas-swat-handlers canvas))
	 (entry (assoc type.modifiers handlers)))
    (if entry
	(set-cdr! entry handler)
	(set-swat-canvas-swat-handlers!
	 canvas (cons (cons type.modifiers handler) handlers)))))

(define-method fix-widget-realize-callback ((canvas <swat-canvas>))
  (call-next-method canvas)
  (realize-options canvas)
  (for-each
    (lambda (type)
      (set-fix-widget-button-handler!
       canvas type
       (named-lambda (canvas-button-handler canvas type button modifiers x y)
	 (%trace "canvas-button-handler "type" "button" "modifiers
		 " "x","y" "canvas)
	 (handle-canvas-event canvas
			      (append! (list type button) modifiers) x y))))
    '(press release double-press triple-press))
  (set-fix-widget-motion-handler!
   canvas
   (named-lambda (canvas-motion-handler canvas modifiers x y)
     (%trace2 "canvas-motion-handler "modifiers" "x","y" "canvas)
     (handle-canvas-event canvas (cons 'motion modifiers) x y))))

(define-class <swat-ink> (<fix-ink>)

  ;; A <swat-group> or #f.
  (group define standard initial-value #f)
  
  ;; An alist of (event-type . modifiers) x SWAT event handler procedures.
  (swat-handlers define standard initial-value '()))

(define (set-swat-ink-handler! item type.modifiers handler)
  (let* ((handlers (swat-ink-swat-handlers item))
	 (entry (assoc type.modifiers handlers)))
    (if entry
	(set-cdr! entry handler)
	(set-swat-ink-swat-handlers!
	 item (cons (cons type.modifiers handler) handlers)))))

(define (item-delete! item)
  (fix-ink-remove! item)
  ;; No hurry here.  Draw callbacks are cut off (above).  Some
  ;; unnecessary drawing damage may occur (if items are changed before
  ;; they lose their link to the drawing), but the redraws will still
  ;; be correct.
  (if (swat-group? item)
      (let unset! ((items (swat-group-items item)))
	(if (pair? items)
	    (let ((i (car items)))
	      (set-fix-ink-drawing! i #f)
	      (if (swat-group? i) (unset! (swat-group-items i)))
	      (unset! (cdr items)))))))

(define (item-move! item dx dy)
  (guarantee fixnum? dx 'item-move!)
  (guarantee fixnum? dy 'item-move!)
  (if (not (and (fix:zero? dx) (fix:zero? dy)))
      (with-glib-lock
       (lambda ()
	 (fix-ink-move! item dx dy)
	 (let ((group (swat-ink-group item)))
	   (if group (recache-group-extent! group)))))))

(define (recache-group-extent! group)
  (let ((extent (fix-ink-extent group))
	(items (swat-group-items group)))
    (if (null? items)
	(set-fix-rect-size! extent 0 0)
	(let ((first (car items))
	      (rest (cdr items)))
	  (copy-fix-rect! extent (fix-ink-extent first))
	  (for-each (lambda (i) (fix-rect-union! extent (fix-ink-extent i)))
		    rest))))
  (let ((group (swat-ink-group group)))
    (if group (recache-group-extent! group))))

(define-class (<swat-group> (constructor ()))
    (<swat-ink>)
  (items define standard initial-value '()))

(define-method fix-ink-draw-callback ((group <swat-group>)
				      widget window cr area)
  (for-each (lambda (ink)
	      (if (fix-ink-in? ink widget area)
		  (fix-ink-draw-callback ink widget window cr area)))
	    (swat-group-items group)))

(define-method fix-ink-move! ((group <swat-group>) dx dy)
  (assert-glib-locked '(fix-ink-move! <swat-group>))
  (let ((extent (fix-ink-extent group)))
    (fix-rect-move! extent dx dy)
    (for-each (lambda (i) (fix-ink-move! i dx dy))
	      (swat-group-items group))))

(define (group-add! group item)
  (with-glib-lock
   (lambda ()
     (let ((items (swat-group-items group))
	   (extent (fix-ink-extent group))
	   (drawing (fix-ink-drawing group)))
       (if (not (eq? drawing (fix-ink-drawing item)))
	   (error "group-add!: wrong drawing:" item group))
       (set-swat-group-items! group (append! items (list item)))
       (if (null? items)
	   (copy-fix-rect! extent (fix-ink-extent item))
	   (fix-rect-union! extent (fix-ink-extent item)))
       (set-fix-drawing-display-list!	;AFTER adding to group extent
	drawing (delq! item (fix-drawing-display-list drawing)))
       (set-swat-ink-group! item group)))))

(define (group-remove! group item)
  (with-glib-lock
   (lambda ()
     (let ((items (swat-group-items group)))
       (define (topmost group)
	 (let ((parent (swat-ink-group group)))
	   (if (not parent) group (topmost parent))))
       (if (memq item items)
	   (begin
	     ;; Move ink to canvas?  Under the topmost group?
	     (fix-drawing-add-ink! (fix-ink-drawing group) item (topmost group))
	     (set-swat-group-items! group (delq! item items))
	     (recache-group-extent! group))
	   (error "No such item:" item group))))))

;;;; Event handling.

;;; Tk provides some fancy massaging of the "raw" input event stream.
;;; The <Double-ButtonPress-1> handler specified by the pole-zero
;;; example is an abbreviation for "<Button-1><Button-1>", according
;;; to the Tk::bind manpage.  ("<Button-..." and "<ButtonPress-..."
;;; are equivalent.)  Such a sequence is matched against the GdkEvent
;;; stream with time and space constraints.
;;;
;;;     The Double, Triple and Quadruple modifiers are a convenience
;;;     for specifying double mouse clicks and other repeated
;;;     events. They cause a particular event pattern to be repeated
;;;     2, 3 or 4 times, and also place a time and space requirement
;;;     on the sequence: for a sequence of events to match a Double,
;;;     Triple or Quadruple pattern, all of the events must occur
;;;     close together in time and without substantial mouse motion in
;;;     between.  For example, <Double-Button-1> is equivalent to
;;;     <Button-1><Button-1> with the extra time and space
;;;     requirement. [Tk::bind(3pm)]
;;;
;;; This implementation uses a preview stage to convert the GdkEvent
;;; stream into "SWAT events".  Thus a Gdk event stream like this:
;;; 
;;;     P-1 R-1 2P-1 R-1
;;; 
;;; 	where P-1 is a GDK_BUTTON_PRESS event whose "button" member = 1,
;;; 	      R-1 is a GDK_BUTTON_RELEASE with button 1,
;;; 	 and 2P-1 is a GDK_2BUTTON_PRESS with button 1
;;;
;;; within certain time and space constraints, generates the event key
;;; 
;;;     (double-press 1)
;;; 
;;; which is looked up among the handlers of canvasitems under the
;;; point.
;;;
;;; Note that the R-1 event must be ignored, not turned into a
;;; (release 1) event, else the pole-zero example breaks in the
;;; <ButtonRelease-1> handler of the newly created object (under the
;;; point).
;;;
;;; The state machine doing the conversion is implemented by the Gtk
;;; event handlers.  Depending on the state of the machine (the event
;;; prefix seen so far) a Gtk event or timeout handler can generate
;;; a SWAT event, or schedule a timeout.
;;;
;;; An example event stream, with small mouse movements, that should
;;; be recognized as a single (double-press 1), no (release 1):
;;; 
;;;     P1 R1 Msmall 2P1 Msmall R1 T0
;;; 
;;;     where Msmall is any number of GDK_MOTION_NOTIFY events that
;;;     stay nearby, and T0 is a timeout scheduled during the handling
;;;     of 2P1.
;;; 
;;; An example event stream that is actually a double-click-drag:
;;;
;;;     P1 Msmall R1 2P1 Msmall Msmall T0 Msmall Msmall R1
;;; 
;;; should produce:
;;; 
;;;     2P1 Msmall Msmall R1
;;;
;;; For now (for pole-zero), the state machine need only ignore button
;;; release (and movement events?) that follow a button press by less
;;; than a fraction of a second, and button presses that are followed
;;; by a double-press within a fraction of a second.
;;;
;;; To handle the latter, it is natural to wait with a timeout for the
;;; double- or triple-press, but sleep-current-thread is inappropriate
;;; in an event handler -- a callback running at "interrupt" level
;;; (within without-interrupts).  Instead, the event handlers queue
;;; "input events" that are read (dequeued), with a timeout, by
;;; swat-thread (which can sleep).

(define (handle-canvas-event canvas event window-x window-y)
  (%trace2 "handle-canvas-event "event" "window-x","window-y" "canvas)
  (let* ((view (fix-layout-view canvas))
	 (x (fix:+ window-x (fix-rect-x view)))
	 (y (fix:+ window-y (fix-rect-y view))))
    (thread-queue/queue-no-hang! swat-input-queue (list event canvas x y))
    #t))

(define swat-input-queue)
(define swat-thread)

(define (initialize-package!)
  (set! swat-input-queue (make-thread-queue 100))
  (set! swat-thread (create-swat-thread))
  (detach-thread swat-thread))

(define (create-swat-thread)
  (create-thread
   #f
   (lambda ()
     (%trace "thread: "(current-thread))

     (let main ()

       (define-integrable (read!)
	 (thread-queue/dequeue! swat-input-queue))

       (define-integrable (peek-until time)
	 (thread-queue/peek-until swat-input-queue time))

       (define-integrable (handle event)
	 (apply handle-event event)
	 (main))

       (let* ((event (read!))
	      (type (caar event)))
	 (cond
	  ((eq? type 'release)
	   (main))			; Ignore EVENT.
	  ((memq type '(press double-press triple-press))
	   (let ((quitn-time (+ 500 (real-time-clock))))
	     (let wait ()
	       (let ((next-event (peek-until quitn-time)))
		 (if (not next-event)
		     (handle event)
		     (let ((next-type (caar next-event)))
		       (cond
			((memq next-type '(double-press triple-press))
			 (main))		; Ignore EVENT, read NEXT-EVENT.
			((eq? next-type 'motion)
			 (read!)		; Ignore NEXT-EVENT,
			 (wait))		; and wait some more.
			(else
			 (handle event)))))))))
	  (else
	   (handle event)))))
     (%trace "thread done")
     (stop-current-thread))
   'swat))

(define (handle-event key canvas x y)
  (or (let* ((handlers (swat-canvas-swat-handlers canvas))
	     (entry (assoc key handlers)))
	(%trace2 "handle-event canvas: "entry)
	(and entry ((cdr entry) canvas x y)))
      (let ((items (pick-list canvas x y)))
	(%trace2 "handle-event pick list: "items)
	(find (lambda (item)
		(let* ((handlers (swat-ink-swat-handlers item))
		       (entry (assoc key handlers)))
		  (if entry
		      (%trace "handle-event "entry" on "item)
		      (%trace2 "handle-event #f on "item))
		  (and entry ((cdr entry) canvas x y))))
	      items))))

(define (pick-list canvas x y)
  (%trace2 "pick-list "x","y" "canvas)
  (let loop ((items (fix-drawing-display-list (fix-layout-drawing canvas)))
	     (picks '()))
    (if (pair? items)
	(loop (cdr items)
	      (let ((item (car items)))
		(if %trace2?
		    (if (not (fix-ink-in-widget? item canvas))
			(%trace2 item" not in "canvas)
			(let ((extent (fix-ink-extent item)))
			  (if (point-in-fix-rect? x y extent)
			      (%trace2 item" ("extent") contains "x","y)
			      (%trace2 item" ("extent") does NOT contain "x","y)
			      ))))
		(if (and (fix-ink-in-widget? item canvas)
			 (point-in-fix-rect? x y (fix-ink-extent item)))
		    (cons item
			  (if (swat-group? item)
			      (append! (loop (swat-group-items item) '())
				       picks)
			      picks))
		    picks)))
	picks)))

(define-class (<swat-line> (constructor ()))
    (<line-ink> <swat-ink>)
  ;; #f, FIRST or SECOND
  (arrow define standard initial-value #f)
  ;; a <swat-polygon>
  (arrow-head define standard initial-value #f))

(define-method fix-ink-move! ((line <swat-line>) dx dy)
  (call-next-method line dx dy)
  (let ((arrow-head (swat-line-arrow-head line)))
    (if arrow-head (fix-ink-move! arrow-head dx dy))))

(define-class (<swat-polygon> (constructor ()))
    (<polygon-ink> <swat-ink>))

(define-class (<swat-rectangle> (constructor ()))
    (<rectangle-ink> <swat-ink>))

(define-class (<swat-oval> (constructor ()))
    (<arc-ink> <swat-ink>))

(define-class (<swat-text> (constructor ()))
    (<simple-text-ink> <swat-ink>)
  (anchor define standard initial-value 'nw
	  modifier set-swat-text-%anchor!))

(define (set-swat-text-anchor! text anchor)
  (if (not (eq? anchor (swat-text-anchor text)))
      (hold-position text
		     (lambda () (set-swat-text-%anchor! text anchor)))))

(define (set-swat-text-text! text string)
  (guarantee string? string 'set-swat-text-text!)
  (if (eq? 'nw (swat-text-anchor text))
      (set-simple-text-ink-text! text string)
      (hold-position text (lambda () (set-simple-text-ink-text! text string)))))

(define-integrable (hold-position text thunk)
  (with-swat-text-pos
   text
   (lambda (old-x old-y)
     (thunk)
     (with-swat-text-pos
      text
      (lambda (new-x new-y)
	(let ((dx (fix:- new-x old-x))
	      (dy (fix:- new-y old-y)))
	  (if (not (and (fix:zero? dx) (fix:zero? dy)))
	      (fix-ink-move! text dx dy))))))))

(define (with-swat-text-pos text receiver)
  ;; Calls RECEIVER with the coordinates of the text's anchor position.
  (let ((anchor (swat-text-anchor text)))
    (with-fix-rect
     (fix-ink-extent text)
     (lambda (x y width height)
       (define-integrable (top) x)
       (define-integrable (left) y)
       (define-integrable (right) (fix:+ x width))
       (define-integrable (bottom) (fix:+ y height))
       (define-integrable (mid-right) (fix:+ x (quotient width 2)))
       (define-integrable (mid-bottom) (fix:+ y (quotient height 2)))
       (case anchor
	 ((CENTER) (receiver (mid-right) (mid-bottom)))
	 ((NE) (receiver (right) (top)))
	 ((SE) (receiver (right) (bottom)))
	 ((SW) (receiver (left) (bottom)))
	 ((NW) (receiver (left) (top)))
	 ((N) (receiver (mid-right) (top)))
	 ((E) (receiver (right) (mid-bottom)))
	 ((S) (receiver (mid-right) (bottom)))
	 ((W) (receiver (left) (mid-bottom)))
	 (else (error "Unexpected text anchor:" anchor text)))))))

;;;; Active Variables

;;; These are tricky.  A variable should, in set-active-variable!,
;;; call any SWAT callback (if set).  Those variables "connected" to a
;;; widget should ALSO frob the widget (e.g. set a GtkLabel's text, or
;;; toggle a check-button,...).
;;;
;;; Some widgets also SET variables, e.g. in a check-button "toggled"
;;; callback.  In these cases, the SWAT callback should be called (if
;;; set), BUT the frobbing of the widget... might be a loop!  Thus an
;;; active variable separates the SWAT callback from the widget
;;; frobbing.  It always does the former, sometimes the latter.

(define-structure (active-variable (constructor make-active-variable ()))

  (value '())

  ;; Could not find anything like SOS slot option "define initpred",
  ;; though there seems to be something there...  Anyway, this seems
  ;; more portable.
  (value-initialized? #f)

  ;; User callback -- a thunk.
  (swat-callback #f)

  ;; Widget frobbing closure, taking one arg: the new value.
  (frob #f))

(define-guarantee active-variable "active-variable")

(define (set-active-variable-callback! variable callback)
  (guarantee-active-variable variable 'set-active-variable-callback!)
  (guarantee-procedure-of-arity callback 0 'set-active-variable-callback!)
  (if (active-variable-swat-callback variable)
      (error "Callback already set:" variable callback))
  (set-active-variable-swat-callback! variable callback))

(define (set-active-variable! variable value)
  (guarantee-active-variable variable 'set-active-variable!)
  (frob-active-variable! variable value)
  (let ((frob (active-variable-frob variable)))
    (if frob (frob value))))

(define-integrable (frob-active-variable! variable value)
  ;; Used by widgets to frob an active variable without getting
  ;; frobbed in return.
  (set-active-variable-value-initialized?! variable #t)
  (set-active-variable-value! variable value)
  (let ((callback (active-variable-swat-callback variable)))
    (if callback (callback))))

;;;; Widget Configuration

;;; Options from the examples:
;;;
;;;     button	-foreground blue
;;;     button	-background yellow
;;;     button	-activebackground red
;;;     button	-text "Push me"
;;;
;;; and from Pole-0:
;;;
;;;     button	-background ,color
;;;     canvas	-background ,color
;;;     label	-background ,color
;;;     button	-font "CourR12"
;;;     label	-font "CourR12"
;;;     label	-relief sunken
;;;     button	-text ,string
;;;     label	-textvariable ,active-variable
;;;     label	-width 13

(define (widget-configure! widget options)
  (let loop ((opts options))
    (if (pair? opts)
	(let ((name (car opts))
	      (value (cadr opts)))
	  (case name
	    ((-foreground) (set-option! widget 'foreground value))
	    ((-background) (set-option! widget 'background value))
	    ((-activeforeground) (set-option! widget 'activeforeground value))
	    ((-activebackground) (set-option! widget 'activebackground value))
	    ((-font) (set-option! widget 'font value))

	    ((-width) (set-width! widget value))
	    ((-text) (set-text! widget value))
	    ((-textvariable) (set-textvariable! widget value))
	    ((-relief) (set-label-relief! widget value))
	    (else (warn "unimplemented configure option:"
			name value widget)))
	  (loop (cddr opts))))))

(define (set-option! widget name spec)
  (let* ((options (swat-widget-options widget))
	 (entry (assq name options)))
    (if entry
	(set-cdr! entry spec)
	(set-swat-widget-options! widget (cons (cons name spec) options))))
  (if (swat-widget-realized? widget) (realize-option widget name spec)))

(define (realize-option widget name spec)
  #;(with-glib-lock
   (lambda ()
     (case name
       ((foreground) (set-gtk-widget-fg-color! widget spec 'normal))
       ((background) (set-gtk-widget-bg-color! widget spec 'normal))
       ((activeforeground) (set-gtk-widget-fg-color! widget spec 'active))
       ((activebackground) (set-gtk-widget-bg-color! widget spec 'active))
       ((font) (set-gtk-widget-font! widget spec))
       (else (warn "Cannot realize widget option:" name spec widget)))))
  (warn "Cannot realize widget option:" name spec widget))

(define (realize-options widget)
  (set-swat-widget-realized?! widget #t)
  (for-each (lambda (option)
	      (realize-option widget (car option) (cdr option)))
	    (swat-widget-options widget)))

(define-integrable (set-width! widget width)
  (if (swat-label? widget)
      (with-glib-lock
       (lambda ()
	 (gtk-label-set-width-chars (gtk-bin-child widget) width)))
      (warn "Unimplemented:" '-width widget)))

(define-generic set-text! (widget string))

(define-method set-text! ((button <swat-button>) string)
  (guarantee string? string '(set-text! <swat-button>))
  (with-glib-lock
   (lambda ()
     (let ((label (gtk-bin-child button)))
       (if (not label)
	   (gtk-container-add button (gtk-label-new string))
	   (gtk-label-set-text label string))))))

(define-method set-text! ((label <swat-label>) string)
  (guarantee string? string '(set-text! <swat-label>))
  (with-glib-lock
   (lambda ()
     (gtk-label-set-text (gtk-bin-child label) string))))

(define-method set-text! ((button <swat-checkbutton>) string)
  (guarantee string? string '(set-text! <swat-checkbutton>))
  (with-glib-lock
   (lambda ()
     (let ((label (gtk-bin-child button)))
       (if (not label)
	   (gtk-container-add button (gtk-label-new string))
	   (gtk-label-set-text label string))))))

(define-generic set-textvariable! (widget active-variable))

(define-method set-textvariable! ((widget <swat-label>) variable)
  (set-active-variable-frob! variable (make-label-frobbery widget))
  (set-active-variable-value! variable
			      (with-glib-lock
			       (lambda ()
				 (gtk-label-get-text (gtk-bin-child widget)))))
  (set-active-variable-value-initialized?! variable #t))

(define (make-label-frobbery label)
  (named-lambda (label-frobbery value)
    (%trace "label-frobbage "label" "value)
    (if (string? value)
	(with-glib-lock
	 (lambda ()
	   (gtk-label-set-text (gtk-bin-child label) value)))
	(warn "Bogus text for swat-label frobbery:" value label))))

;;;; Canvas Item Configuration

;;; Options from pole-zero:
;;;
;;;     text	-anchor sw
;;;     text	-anchor e
;;;     line	-arrow last
;;;     line	-arrow first
;;;     group	-fill ,color
;;;     oval	-fill ,color
;;;     text	-font "-adobe-symbol-...-fontspecific"
;;;     oval	-outline "gray"
;;;     text	-text "p"
;;;     group	-width 2
;;;     oval	-width 2

(define (item-configure! item options)
  (with-glib-lock
   (lambda ()
     (let loop ((opts options))
       (if (pair? opts)
	   (let ((name (car opts))
		 (value (cadr opts)))
	     (set-item-option! item name value)
	     (loop (cddr opts))))))))

(define (set-item-option! item name value)
  (if (swat-group? item)
      (for-each (lambda (i) (set-item-option! i name value))
		(swat-group-items item))
      (case name
	((-anchor) (set-item-anchor! item value))
	((-arrow) (set-item-arrow! item value))
	((-fill) (set-item-fill! item value))
	((-font) (set-item-font! item value))
	((-outline) (set-item-outline! item value))
	((-text) (set-item-text! item value))
	((-width) (set-item-width! item value))
	(else (error "Unimplemented option:" name value item)))))

(define-generic set-item-anchor! (item value))
(define-method set-item-anchor! ((text <swat-text>) value)
  (if (not (memq value '(N S E W NE SE SW NW)))
      (error:wrong-type-argument value "an anchor direction" 'set-item-anchor!))
  (set-swat-text-anchor! text value))

(define-generic set-item-arrow! (item end))
(define-method set-item-arrow! ((line <swat-line>) end)
  (if (not (memq end '(#F FIRST LAST)))
      (error:wrong-type-argument end "#F, 'FIRST or 'LAST" 'SET-ITEM-ARROW!))
  (let ((head (swat-line-arrow-head line)))
    (set-swat-line-arrow! line end)
    (cond ((and end (not head))
	   (set-swat-line-arrow-head! line (make-arrow-head line end)))
	  ((and (not end) head)
	   (fix-ink-remove! head)
	   (set-swat-line-arrow-head! line #f)))))

(define (make-arrow-head line end)
  (let ((head (with-glib-lock make-swat-polygon)))
    (set-polygon-ink-fill-color! head (let ((c (line-ink-color line)))
					(if (null? c) "black" c)))
    (update-arrow-head head line end)
    (fix-drawing-add-ink! (fix-ink-drawing line) head line)
    (set-swat-line-arrow-head! line head)))

(define (update-arrow-head ink line end)
  (set-polygon-ink!
   ink
   (with-fix-rect
    (line-ink-vector line)
    (lambda (startx starty dx dy)
      (with-glib-lock
       (lambda ()
	 (let ((matrix (cairo-rotation-matrix
			(flo:atan2 (flo:- 0. (->flonum dy)) (->flonum dx)))))
	   (cairo-matrix-scale!
	    matrix
	    (->flonum (let ((w (line-ink-width line)))
			(if (null? w) 1 w))))
	   (cairo-matrix-translate!
	    matrix
	    (->flonum (if (eq? end 'FIRST) startx (+ startx dx)))
	    (->flonum (if (eq? end 'FIRST) starty (+ starty dy))))
	   (let ((pt (cairo-point 0. 0.)))
	     (map (lambda (pair)
		    (set-x! pt (->flonum (car pair)))
		    (set-y! pt (->flonum (cdr pair)))
		    (cairo-transform! pt matrix)
		    (cons (round->exact (x pt))
			  (round->exact (y pt))))
		  '((0 . 0) (-10 . 2) (-10 . -2) (0 . 0)))))))))))

(define-generic set-item-fill! (item color))
(define-method set-item-fill! ((item <swat-line>) color)
  (let ((head (swat-line-arrow-head item)))
    (if head (set-polygon-ink-color! head color)))
  (set-line-ink-color! item color))
(define-method set-item-fill! ((item <swat-oval>) color)
  (set-arc-ink-fill-color! item color))
(define-method set-item-fill! ((item <swat-rectangle>) color)
  (set-rectangle-ink-fill-color! item color))
(define-method set-item-fill! ((item <swat-text>) color)
  (set-text-ink-color! item color))

(define-generic set-item-font! (item font))
(define-method set-item-font! ((item <swat-text>) font)
  (set-simple-text-ink-font! item font))

(define-generic set-item-outline! (item value))
(define-method set-item-outline! ((item <swat-oval>) color)
  (set-arc-ink-color! item color))
(define-method set-item-outline! ((item <swat-rectangle>) color)
  (set-rectangle-ink-color! item color))

(define-generic set-item-text! (item value))
(define-method set-item-text! ((text <swat-text>) string)
  (guarantee string? string '(set-item-text! <swat-text>))
  (set-swat-text-text! text string))

(define-generic set-item-width! (item value))
(define-method set-item-width! ((item <swat-line>) width)
  (let ((head (swat-line-arrow-head item)))
    (if head (set-polygon-ink-width! head width)))
  (set-line-ink-width! item width))
(define-method set-item-width! ((item <swat-oval>) width)
  (guarantee positive-fixnum? width '(set-item-width! <swat-oval>))
  (set-arc-ink-width! item width))
(define-method set-item-width! ((item <swat-rectangle>) width)
  (guarantee positive-fixnum? width '(set-item-width! <swat-rectangle>))
  (set-rectangle-ink-width! item width))

;;;; SWAT Interface

;;; Here are the procedures exported to the (swat) package.  They are
;;; organized by their original SWAT files, listed in the order they
;;; are loaded.

;;; * control-floating-errors
;;;
;;; LAP-generating floating-exception-flag-frobing Nerds Gone Wild.

;;; * structures
;;;
;;; Definitions of structures in Tk.

;;; * structures2
;;;
;;; And some more.

;;; * generics
;;;
;;; Generic operations:
;;;   SET-CONTEXT!, assign-screen-area!,
;;;   assign-drawing-surface!, handle-event,
;;;   get-desired-size, get-desired-size, EVENT-WITHIN?

(define (add-child! object child . others)
  (if (null? others)
      (with-glib-lock (lambda () (gtk-container-add object child)))
      (error "unimplemented")))

(define (remove-child! object child)
  (with-glib-lock (lambda () (gtk-container-remove object child))))

;;; Ask-widget in the examples:
;;;
;;; (ask-widget button-widget '(configure -foreground blue))
;;;
;;; and in pole-zero.scm:
;;;
;;; (ask-widget oval-item `(configure -outline ,zero-color ...))
;;; (ask-widget group-item `(configure -fill ,pole-color -width 2))
;;; (ask-widget button-widget `(configure -text ,(car (list-ref to-switch i))))
;;; (ask-widget label-widget `(configure -width 13 ...))
;;; (ask-widget canvas-widget `(configure -background ,canvas-color))
;;; (ask-widget line-item '(configure -arrow last))
;;; (ask-widget text-item '(configure -anchor sw ...))
;;; (ask-widget oval-item `(move ,(- x last-x) ,(- y last-y)))
;;; (ask-widget group-item `(move ,(- x last-x) ,(- y last-y)))
;;; (ask-widget oval-item '(delete))
;;; (ask-widget group-item '(delete))
;;; (ask-widget canvas-widget '(delete all))

(define (ask-widget object command)
  ;; Dispatch on command name and object type (widget or canvas item).
  (case (car command)
    ((CONFIGURE)
     (cond ((swat-widget? object)
	    (widget-configure! object (cdr command)))
	   ((swat-ink? object)
	    (item-configure! object (cdr command)))
	   (else (warn "Non-SWAT widget:" object))))
    ((MOVE)
     (cond ((and (swat-ink? object)
		 (= 3 (length command)))
	    (item-move! object (cadr command) (caddr command)))
	   (else
	    (warn "Unexpected move command:" object command))))
    ((DELETE)
     (cond ((and (equal? command '(DELETE ALL))
		 (swat-canvas? object))
	    (delete-all! object))
	   ((and (equal? command '(DELETE))
		 (swat-ink? object))
	    (item-delete! object))
	   (else
	    (warn "Unexpected delete command:" object command))))
    (else (warn "Unimplemented ask-widget command:" object command))))

(define-generic add-event-handler! (object event-type handler . substitutions))
;;; Examples: (add-event-handler! button "<Enter>" (lambda () ...))
;;; Pole-0: (add-event-handler! item-group "<ButtonPress-1>" (lambda (x y) ...) "%x" "%y")
;;; Pole-0: (add-event-handler! item-group "<ButtonPress-1>" (lambda (x) ...) "%x")
;;; Pole-0: (add-event-handler! item-group "<ButtonRelease-1>" (lambda () ...))
;;; Pole-0: (add-event-handler! item-group "<B1-Motion>" (lambda (x y) ...) "%x" "%y")
;;; Pole-0: (add-event-handler! item-group "<B1-Motion>" (lambda (x) ...) "%x")

(define-method add-event-handler! ((button <swat-button>)
				   event-type handler . substitutions)
  (cond ((and (string=? event-type "<Enter>")
	      (null? substitutions))
	 (guarantee-procedure-of-arity handler 0
				       'add-event-handler!-<swat-button>)
	 (set-gtk-button-clicked-callback!
	  button (lambda (button) (declare (ignore button)) (handler))))
	(else
	 (warn "Unimplemented:" '(add-event-handler! <swat-button>)
	       button event-type handler substitutions))))

(define-method add-event-handler! ((canvas <swat-canvas>)
				   event-type handler . substitutions)

  (define (unimplemented)
    (warn "Unimplemented:" '(add-event-handler! <swat-canvas>)
	  canvas event-type handler substitutions))

  (cond
   ((and (string=? event-type "<Double-ButtonPress-1>")
	 (equal? substitutions '("%x" "%y")))
    (guarantee-procedure-of-arity handler 2 '(add-event-handler! <swat-canvas>))
    (set-swat-canvas-handler! canvas '(double-press 1)
			      (lambda (canvas x y)
				(declare (ignore canvas))
				(handler x y))))
   (else (unimplemented))))

(define-method add-event-handler! ((item <swat-ink>)
				   event-type handler . substitutions)

  (define (unimplemented)
    (warn "Unimplemented:" '(add-event-handler! <swat-ink>)
	  item event-type handler substitutions))

  (cond
   ((string=? event-type "<ButtonPress-1>")
    (cond
     ((equal? substitutions '("%x" "%y"))
      (guarantee-procedure-of-arity handler 2 '(add-event-handler! <swat-ink> "<ButtonPress-1>" "%x" "%y"))
      (set-swat-ink-handler! item '(press 1)
			     (lambda (item x y)
			       (declare (ignore item))
			       (handler x y))))
     ((equal? substitutions '("%x"))
      (guarantee-procedure-of-arity handler 1 '(add-event-handler! <swat-ink> "<ButtonPress-1>" "%x"))
      (set-swat-ink-handler! item '(press 1)
			     (lambda (item x y)
			       (declare (ignore item y))
			       (handler x))))
     (else (unimplemented))))
   ((string=? event-type "<ButtonRelease-1>")
    (cond
     ((equal? substitutions '())
      (guarantee-procedure-of-arity handler 0 '(add-event-handler! <swat-ink> "<ButtonRelease-1>"))
      (set-swat-ink-handler! item '(release 1 button1)
			     (named-lambda (release1 item x y)
			       (declare (ignore item x y))
			       (handler))))
     (else (unimplemented))))
   ((string=? event-type "<B1-Motion>")
    (cond
     ((equal? substitutions '("%x" "%y"))
      (guarantee-procedure-of-arity handler 2 '(add-event-handler! <swat-ink> "<B1-Motion>" "%x" "%y"))
      (set-swat-ink-handler! item '(motion button1)
			     (named-lambda (motion-button1-x-y item x y)
			       (declare (ignore item))
			       (handler x y))))
     ((equal? substitutions '("%x"))
      (guarantee-procedure-of-arity handler 1 '(add-event-handler! <swat-ink> "<B1-Motion>" "%x"))
      (set-swat-ink-handler! item '(motion button1)
			     (named-lambda (motion-button1-x item x y)
			       (declare (ignore item y))
			       (handler x))))
     (else (unimplemented))))
   (else (unimplemented))))

(define-generic set-callback! (object callback))
;;; Examples: (set-callback! button (lambda () ...))

(define-method set-callback! ((object <swat-button>) callback)
  (guarantee-procedure-of-arity callback 0 '(set-callback! <swat-button>))
  (set-gtk-button-clicked-callback!
   object (lambda (button) (declare (ignore button)) (callback))))

(define-method set-callback! ((object <swat-checkbutton>) callback)
  (guarantee-procedure-of-arity callback 0 '(set-callback! <swat-checkbutton>))
  (set-swat-checkbutton-swat-callback! object callback))

(define-method set-callback! ((object rtd:active-variable) callback)
  (guarantee-procedure-of-arity callback 1 '(set-callback! active-variable))
  (set-active-variable-callback! object callback))

;;; * uitk
;;;
;;; Geometry, Events, Contexts, rectangle=, UIObj-protect-from-gc!...

#;(define (create-default-context name display)
  name display
  unspecific)

(define (on-death! object reason thunk)
  (let ((old (swat-widget-on-death object)))
    (if old (warn "Swat-widget already has an on-death! callback:"
		  object reason thunk old)))
  (set-swat-widget-on-death! object (cons reason thunk)))

;;; * xlibCONSTANTS
;;;
;;; xlib constants: enums, masks,

;;; * mit-xlib
;;;
;;; The SCXL library for interfacing with X.  E.g. XClearArea,
;;; initialize-scxl!.

;;; * tk-mit
;;;
;;; tk-init, %tkKillApplication, TK Callback handling.

;;; * mit-xhooks
;;;
;;; UITK main loop, make-uitk-thread, when-idle, after-delay,
;;; ClearArea, initialize-uitk!.

(define (after-delay seconds thunk)
  (guarantee index-fixnum? seconds 'after-delay)
  (guarantee-procedure-of-arity thunk 0 'after-delay)
  (detach-thread
   (create-thread
    #f
    (lambda ()
      (%trace "after-delay "seconds", sleeping "(current-thread))
      (sleep-current-thread (* seconds 1000))
      (thunk)
      (stop-current-thread)))))

;;; * widget-mit
;;;
;;; tk-make-menu, initialize-mit-widgets!.

;;; * baseobj
;;;
;;; Basic objects for the Scheme User Interface Tool Kit.
;;; E.g. make-application, swat-open, swat-close.

(define (swat-open . options)
  (let loop ((options options)
	     (objects '()))
    (if (pair? options)
	(if (eq? '-title (car options))
	    (if (and (pair? (cdr options)) (string? (cadr options)))
		(let ((window (%open (reverse! objects) (cadr options))))
		  (cons window (loop (cddr options) '())))
		(error "No title after -title keyword:" options))
	    (loop (cdr options) (cons (car options) objects)))
	(if (null? objects)
	    '()
	    (list (%open (reverse! objects) "SWAT"))))))

(define (%open children title)
  (with-glib-lock
   (lambda ()
     (let ((window (gtk-window-new 'toplevel))
	   (grid (gtk-grid-new)))
       (gtk-grid-set-row-spacing grid 5)
       (gtk-grid-set-column-spacing grid 5)
       (gtk-orientable-set-orientation grid 'horizontal)
       (for-each (lambda (child) (gtk-container-add grid child)) children)
       (gtk-window-set-title window title)
       (gtk-container-add window grid)
       (gtk-widget-show-all window)
       window))))

(define (swat-close child)
  (let ((parent (gtk-widget-parent child)))
    (if parent
	(swat-close parent)
	(or (with-glib-lock
	     (lambda ()
	       (and (gtk-window? child)
		    (begin
		      (gtk-widget-destroy child)
		      #t))))
	    (error "unexpected top-level widget" child)))))

;;; * widget
;;;
;;; TkWidget-add-event-handler!, TkWidget-ask-widget, make-listbox,
;;; make-active-variable, initialize-widgets!.

(define (make-button #!optional options)
  (let ((button (with-glib-lock make-swat-button)))
    (if (not (default-object? options))
	(widget-configure! button options))
    button))

(define (make-label #!optional options)
  (let ((label (with-glib-lock make-swat-label)))
    (if (not (default-object? options))
	(widget-configure! label options))
    label))

(define (make-checkbutton options)
  (let ((button (with-glib-lock make-swat-checkbutton)))
    (set-gtk-check-button-toggled-callback! button checkbutton-toggled-callback)
    (let ((active (find-option options '-variable #f)))
      (if active
	  (begin
	    (set-swat-checkbutton-swat-variable! button active)
	    (set-active-variable-frob!
	     active (make-checkbutton-frobbery button)))))
    (widget-configure! button (delete-option! '-variable options))
    button))

(define (make-checkbutton-frobbery button)
  (named-lambda (checkbutton-frobbery value)
    (%trace "checkbutton-frobbery: setting "button" to "value)
    (with-glib-lock (lambda () (gtk-check-button-set-active button value)))))

(define (checkbutton-toggled-callback button)
  (assert-glib-locked 'checkbutton-toggled-callback)
  (let ((variable (swat-checkbutton-swat-variable button))
	(callback (swat-checkbutton-swat-callback button)))
    (if (or variable callback)
	(let ((state (gtk-check-button-get-active button)))
	  (if variable
	      (begin
		(%trace "checkbutton-toggled-callback:"
		        " setting "variable" to "state" for "button)
		(frob-active-variable! variable state)))
	  (if callback
	      (begin
		(%trace "checkbutton-toggled-callback:"
		        " calling "callback" for "button)
		(callback))))
	(%trace "checkbutton-toggled-callback: noop"))))

(define (checkbutton-variable-on? active)
  (active-variable-value active))

;;; * geometry
;;;
;;; make-hbox, make-array-box.

(define (make-hbox . kids)
  (with-glib-lock
   (lambda ()
     (let ((grid (gtk-grid-new)))
       (gtk-orientable-set-orientation grid 'horizontal)
       (for-each (lambda (kid) (gtk-container-add grid kid)) kids)
       grid))))

(define (make-vbox . kids)
  (with-glib-lock
   (lambda ()
     (let ((grid (gtk-grid-new)))
       (gtk-orientable-set-orientation grid 'vertical)
       (for-each (lambda (kid) (gtk-container-add grid kid)) kids)
       grid))))

(define (box-children box)
  (gtk-container-children box))

;;; * simple
;;;
;;; make-shape, make-rect, make-oval, make-line.

;;; * canvas
;;;
;;; make-scrollable-canvas, ..., make-bitmap-on-canvas...

(define (make-canvas options)
  (let ((width (find-option options '-width #f))
	(height (find-option options '-height #f)))
    (let ((canvas (with-glib-lock (lambda () (make-swat-canvas width height)))))
      (%trace "make-canvas "options": configuring "canvas)
      (widget-configure! canvas (delete-options! '(-width -height) options))
      (%trace "make-canvas "options": "canvas)
      canvas)))

(define (make-canvas-item-group canvas items)
  (let ((group (make-swat-group)))
    (fix-drawing-add-ink! (fix-layout-drawing canvas) group)
    (for-each (lambda (item) (group-add! group item)) items)
    group))

(define (delete-all! canvas)
  (let ((drawing (fix-layout-drawing canvas)))
    (for-each fix-ink-remove! (fix-drawing-display-list drawing))))

(define (make-line-on-canvas canvas x1 y1 x2 y2)
  (guarantee fixnum? x1 'make-line-on-canvas)
  (guarantee fixnum? y1 'make-line-on-canvas)
  (guarantee fixnum? x2 'make-line-on-canvas)
  (guarantee fixnum? y2 'make-line-on-canvas)
  (let ((item (make-swat-line)))
    (set-line-ink! item x1 y1 x2 y2)
    (fix-drawing-add-ink! (fix-layout-drawing canvas) item)
    item))

(define (make-rectangle-on-canvas canvas x y width height)
  (guarantee fixnum? x 'make-rectangle-on-canvas)
  (guarantee fixnum? y 'make-rectangle-on-canvas)
  (guarantee positive-fixnum? width 'make-rectangle-on-canvas)
  (guarantee positive-fixnum? height 'make-rectangle-on-canvas)
  (let ((item (make-swat-rectangle)))
    (set-rectangle-ink! item x y width height)
    (fix-drawing-add-ink! (fix-layout-drawing canvas) item)
    item))

(define (make-oval-on-canvas canvas x1 y1 x2 y2)
  (guarantee fixnum? x1 'make-oval-on-canvas)
  (guarantee fixnum? y1 'make-oval-on-canvas)
  (guarantee fixnum? x2 'make-oval-on-canvas)
  (guarantee fixnum? y2 'make-oval-on-canvas)
  (let ((x (fix:min x1 x2))
	(y (fix:min y1 y2))
	(width (fix:abs (fix:- x2 x1)))
	(height (fix:abs (fix:- y2 y1)))
	(item (make-swat-oval)))
    (set-arc-ink! item x y width height)
    (set-arc-ink-start-angle! item 0)
    (set-arc-ink-sweep-angle! item 360)
    (fix-drawing-add-ink! (fix-layout-drawing canvas) item)
    item))

(define (make-text-on-canvas canvas x y options)
  (let ((text (find-option options '-text ""))
	(anchor (find-option options '-anchor 'center))
	(ink (make-swat-text)))
    (guarantee string? text 'make-text-on-canvas)
    (set-simple-text-ink-text! ink canvas text)
    (set-text-ink-position! ink x y)
    (set-swat-text-anchor! ink anchor)
    (fix-drawing-add-ink! (fix-layout-drawing canvas) ink)
    (item-configure! ink (delete-options! '(-text -anchor) options))
    ink))

;;; * menu
;;;
;;; find-menu-record, delete-menuitem! add-to-menu.

;;; * text
;;;
;;; make-scrollable-text, make-text-tag, texttag-add-event-handler!

(define (find-option options name default)
  (let loop ((opts options))
    (if (pair? opts)
	(if (eq? (car opts) name)
	    (cadr opts)
	    (loop (cddr opts))
	    ;; If multiple values can follow an option keyword:
	    #;(let collect ((opts (cdr opts))
			  (sgra '()))
	      (if (or (null? opts)
		      (and (symbol? (car opts))
			   (char=? #\-
				   (string-ref (symbol-name (car opts)) 0))))
		  ;; Return a list only if multiple values followed.
		  (if (and (not (null? sgra))
			   (null? (cdr sgra)))
		      (car sgra)
		      (reverse! sgra))
		  (collect (cdr opts) (cons (car opts) sgra))))
	    )
	default)))

(define (option-names options)
  (let loop ((opts options))
    (if (null? opts)
	'()
	(cons (car opts) (loop (cddr opts))))))

(define (delete-options! names options)
  (let loop ((names names)
	     (options options))
    (if (pair? names)
	(loop (cdr names)
	      (delete-option! (car names) options))
	options)))

(define (delete-option! name options)

  (define (trim options)
    (if (pair? options)
	(if (eq? (car options) name)
	    (cddr options)
	    (begin
	      (splice! (cdr options) (cddr options))
	      options))
	'()))

  (define (splice! previous options)
    (if (pair? options)
	(if (eq? name (car options))
	    (set-cdr! previous (cddr options))
	    (splice! (cdr options) (cddr options)))))

  (trim options))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace? (outf-error "; swat: " ARGS ... "\n")))))

(define %trace2? #f)

(define-syntax %trace2
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace2? (outf-error "; swat: " ARGS ... "\n")))))

(initialize-package!)